from sklearn.preprocessing import OneHotEncoder
import tensorflow as tf
from sklearn.mixture import BayesianGaussianMixture
from teras.utils import tf_random_choice
import numpy as np
import pandas as pd
from teras.preprocessing.base import BaseDataTransformer
from teras.utils.types import FeaturesNamesType


class ModeSpecificNormalization:
    """
    Mode Specific Normalization as proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        numerical_features: ``List[str]``
            List of numerical features names.
            In the case of ndarray, pass a list of numerical column indices.

        max_clusters: ``int``, default 10,
            Maximum clusters

        std_multiplier: ``int``, default 4,
            Multiplies the standard deviation in the normalization.
            Defaults to 4 as proposed in the paper.

        weight_threshold: ``int``, default 0.005,
            Taken from the official implementation.
            The minimum value a component weight can take to be considered a valid component.
            `weights_` under this value will be ignored.

        covariance_type: ``str``,
            Parameter for the ``GaussianMixtureModel`` class of sklearn

        weight_concentration_prior_type: ``str``, default "dirichlet_process",
            Parameter for the ``GaussianMixtureModel`` class of sklearn

        weight_concentration_prior: ``float``, default 0.01,
            Parameter for the GaussianMixtureModel class of sklearn
    """
    def __init__(self,
                 numerical_features: FeaturesNamesType = None,
                 max_clusters: int = 10,
                 std_multiplier: int = 4,
                 weight_threshold: float = 0.,
                 covariance_type: str = "full",
                 weight_concentration_prior_type: str = "dirichlet_process",
                 weight_concentration_prior: float = 0.001):
        self.numerical_features = numerical_features
        self.max_clusters = max_clusters
        self.std_multiplier = std_multiplier
        self.weight_threshold = weight_threshold
        self.covariance_type = covariance_type
        self.weight_concentration_prior_type = weight_concentration_prior_type
        self.weight_concentration_prior = weight_concentration_prior

        # Features meta-data dictionary will contain all the information about a feature
        # such as selected clusters indices, number of valid clusters, clusters means & stds.
        #
        # 1. `Clusters indices` will be used in the transform method
        #       to create the one hot vector B(i,j) where B stands for Beta, for each value c(i,j)
        #       where c(i,j) is the jth value in the ith numerical feature.
        #       as proposed in the paper in the steps to apply mode specific normalization method on page 3-4.
        # 2. `Number of valid clusters` will be used in the transform method when one-hotting
        # 3. `means` and `standard deviations` will be used in transform
        #       step to normalize the value c(i,j) to create a(i,j) where a stands for alpha.
        self.metadata = {}

        self.bay_guass_mix = BayesianGaussianMixture(n_components=self.max_clusters,
                                                     covariance_type=self.covariance_type,
                                                     weight_concentration_prior_type=self.weight_concentration_prior_type,
                                                     weight_concentration_prior=self.weight_concentration_prior)

        self.fitted = False

    def fit(self, x):
        """
        Args:
            x: A pandas DataFrame or numpy ndarray
        """
        relative_indices_all = []
        num_valid_clusters_all = []
        relative_index = 0
        for feature_name in self.numerical_features:
            if isinstance(x, pd.DataFrame):
                feature = x[feature_name].values.reshape(-1, 1)
            elif isinstance(x, np.ndarray):
                feature = x[:, feature_name].reshape(-1, 1)
            else:
                raise ValueError(f"`x` must be either a pandas DataFrame or numpy ndarray. "
                                 f"{type(x)} was given.")

            self.metadata[feature_name] = {}
            self.bay_guass_mix.fit(feature)
            # The authors use a weight threshold to filter out components in their implementation.
            # For consistency's sake, we're going to use this idea but with slight modification.
            # Reference to the official implementation: https://github.com/sdv-dev/CTGAN
            valid_clusters_indicator = self.bay_guass_mix.weights_ > self.weight_threshold
            # Compute probability of coming from each cluster for each value in the given (numerical) feature.
            clusters_probs = self.bay_guass_mix.predict_proba(feature)
            # Filter out the "invalid" clusters
            clusters_probs = clusters_probs[:, valid_clusters_indicator]
            clusters_probs[np.isnan(clusters_probs)] = 1e-10
            # Normalize probabilities to sum up to 1 for each row
            clusters_probs /= np.sum(clusters_probs, axis=1, keepdims=True)

            # Sample/Select a cluster for each value based on the given clusters probabilities array for that value
            def random_choice(probs_array):
                probs_array[np.isnan(probs_array)] = 1e-10
                probs_array /= np.sum(probs_array)
                return np.random.choice(np.arange(len(probs_array)), p=probs_array)

            selected_clusters_indices = np.apply_along_axis(
                random_choice,
                axis=1,
                arr=clusters_probs)

            # To create one-hot component, we'll store the selected clusters indices
            # and the number of valid clusters
            num_valid_clusters = sum(valid_clusters_indicator)
            self.metadata[feature_name]['selected_clusters_indices'] = selected_clusters_indices
            self.metadata[feature_name]['num_valid_clusters'] = num_valid_clusters

            relative_indices_all.append(relative_index)
            # The 1 is for the alpha feature, since each numerical feature
            # gets transformed into alpha + betas where,
            # |alpha| = 1 and |betas| = num_valid_clusters
            relative_index += (1 + num_valid_clusters)
            num_valid_clusters_all.append(num_valid_clusters)

            # Use the selected clusters to normalize the values.
            # To normalize, we need the means and standard deviations
            # Means
            clusters_means = self.bay_guass_mix.means_.squeeze()[valid_clusters_indicator]
            self.metadata[feature_name]['clusters_means'] = clusters_means
            # Standard Deviations
            clusters_stds = np.sqrt(self.bay_guass_mix.covariances_).squeeze()[valid_clusters_indicator]
            self.metadata[feature_name]['clusters_stds'] = clusters_stds

        self.metadata["relative_indices_all"] = np.array(relative_indices_all)
        self.metadata["num_valid_clusters_all"] = num_valid_clusters_all
        self.fitted = True

    def transform(self, x):
        """
        Args:
            x: A pandas DataFrame
        Returns:
            A numpy ndarray of transformed numerical data
        """
        # Contain the normalized numerical features
        x_cont_normalized = []
        for feature_name in self.numerical_features:
            selected_clusters_indices = self.metadata[feature_name]['selected_clusters_indices']
            num_valid_clusters = self.metadata[feature_name]['num_valid_clusters']
            # One hot components for all values in the feature
            # we borrow the beta notation from the paper
            # for clarity and understanding's sake.
            betas = np.eye(num_valid_clusters)[selected_clusters_indices]

            # Normalizing
            means = self.metadata[feature_name]['clusters_means']
            stds = self.metadata[feature_name]['clusters_stds']
            if isinstance(x, pd.DataFrame):
                feature = x[feature_name].values
            elif isinstance(x, np.ndarray):
                feature = x[:, feature_name]
            else:
                raise ValueError(f"`x` must be either a pandas DataFrame or numpy ndarray. "
                                 f"{type(x)} was given.")
            means = means[selected_clusters_indices]
            stds = stds[selected_clusters_indices]
            alphas = (feature - means) / (self.std_multiplier * stds)
            alphas = np.expand_dims(alphas, 1)

            normalized_feature = np.concatenate([alphas, betas], axis=1)
            x_cont_normalized.append(normalized_feature)

        x_cont_normalized = np.concatenate(x_cont_normalized, axis=1)
        return x_cont_normalized

    def fit_transform(self, x):
        """
        Fits and transforms x.
        Returns:
            A normalized copy of x.
        """
        self.fit(x)
        return self.transform(x)

    def reverse_transform(self, x_normalized):
        """
        Args:
            x: DataFrame or n-dimensional numpy array
        Returns:
            Convert the normalized features to original values,
            effectively reversing the transformation
        """
        x = x_normalized.copy()
        for feature_name in self.numerical_features:
            means = self.features_clusters_means[feature_name]
            stds = self.features_clusters_stds[feature_name]
            if isinstance(x_normalized, pd.DataFrame):
                normalized_feature = x_normalized[feature_name].values
                x[feature_name] = normalized_feature * (self.std_multiplier * stds) + means
            elif isinstance(x_normalized, np.ndarray):
                # todo delete this part and just stick with pandas dataframe
                normalized_feature = x_normalized[:, feature_name]
                x[:, feature_name] = normalized_feature * (self.std_multiplier * stds) + means
            else:
                raise ValueError(f"`x_normalized` must be either a pandas DataFrame or numpy ndarray. "
                                 f"{type(x_normalized)} was given.")
        return x


class CTGANDataTransformer(BaseDataTransformer):
    """
    Data Transformation class based on the data transformation
    in the official CTGAN paper and implementation.

    Reference(s):
        https://arxiv.org/abs/1907.00503
        https://github.com/sdv-dev/CTGAN/

    Args:
        categorical_features: ``List[str]``,
            List of categorical features names in the dataset.

        numerical_features: ``List[str]``,
            List of numerical features names in the dataset.

        max_clusters: ``int``, default 10,
            Maximum Number of clusters to use in ``ModeSpecificNormalization``

        std_multiplier: ``int``, default 4,
            Multiplies the standard deviation in the normalization.

        weight_threshold: ``float``, default 0.005,
            The minimum value a component weight can take to be considered a valid component.
            `weights_` under this value will be ignored.
            (Taken from the official implementation.)

        covariance_type: ``str``, default "full",
            Parameter for the ``GaussianMixtureModel`` class of sklearn.

        weight_concentration_prior_type: ``str``, default "dirichlet_process",
            Parameter for the ``GaussianMixtureModel`` class of sklearn

        weight_concentration_prior: ``float``, default 0.001,
            Parameter for the ``GaussianMixtureModel`` class of sklearn.
    """
    def __init__(self,
                 numerical_features: FeaturesNamesType = None,
                 categorical_features: FeaturesNamesType = None,
                 max_clusters: int = 10,
                 std_multiplier: int = 4,
                 weight_threshold: float = 0.005,
                 covariance_type: str = "full",
                 weight_concentration_prior_type: str = "dirichlet_process",
                 weight_concentration_prior: float = 0.001
                 ):
        self.numerical_features = numerical_features if numerical_features else []
        self.categorical_features = categorical_features if categorical_features else []
        self.max_clusters = max_clusters
        self.std_multiplier = std_multiplier
        self.weight_threshold = weight_threshold
        self.covariance_type = covariance_type
        self.weight_concentration_prior_type = weight_concentration_prior_type
        self.weight_concentration_prior = weight_concentration_prior

        self.num_categorical_features = len(categorical_features)
        self.num_numerical_features = len(numerical_features)

        self.mode_specific_normalizer = None
        if self.num_numerical_features > 0:
            self.mode_specific_normalizer = ModeSpecificNormalization(
                numerical_features=self.numerical_features,
                max_clusters=self.max_clusters,
                std_multiplier=self.std_multiplier,
                weight_threshold=self.weight_threshold,
                covariance_type=self.covariance_type,
                weight_concentration_prior_type=self.weight_concentration_prior_type,
                weight_concentration_prior=self.weight_concentration_prior)
        self.categorical_values_probs = dict()
        self.one_hot_enc = OneHotEncoder()
        self.metadata = dict()
        self.metadata["categorical"] = dict()
        self.metadata["numerical"] = dict()

    def transform_numerical_data(self, x):
        return self.mode_specific_normalizer.fit_transform(x)

    def transform_categorical_data(self, x):
        # To speedup computation of conditional vector down the road,
        # we assign a relative index to each feature. For instance,
        # Given three categorical columns Gender(2 categories), City (4 categories) and EconomicClass (5 categories)
        # Relative indexes will be calculated as below:
        #   gender_relative_index: 0
        #   city_relative_index: gender_relative_index + num_categories_in_gender => 0 + 2 = 2
        #   economic_class_relative_index: city_relative_index + num_categories_in_city => 2 + 4 = 6

        categorical_metadata = dict()
        # NOTE: The purpose of using these lists is that, this way we'll later be able to access
        # metadata for multiple features at once using their indices rather than names
        # which would've been required in case of a dict and would have been less efficient
        relative_indices_all = []
        num_categories_all = []
        # For every feature we'll compute the probabilities over the range of values based on their freqs
        # and then append that probabilities array to the following mother array
        categories_probs_all = []
        # A nested list where each element corresponds to the list of categories in the feature
        categories_all = []
        feature_relative_index = 0
        for feature_name in self.categorical_features:
            num_categories = x[feature_name].nunique()
            num_categories_all.append(num_categories)
            relative_indices_all.append(feature_relative_index)
            feature_relative_index += num_categories

            log_freqs = x[feature_name].value_counts().apply(np.log)
            categories_probs_dict = log_freqs.to_dict()
            # To overcome the floating point precision issue which causes probabilities to not sum up to 1
            # and resultantly causes error in np.random.choice method,
            # we round the probabilities to 7 decimal points
            probs = np.around(np.array(list(categories_probs_dict.values())), 7)
            # Normalizing so all probs sum up to 1
            probs = probs / np.sum(probs)
            categories, categories_probs = list(categories_probs_dict.keys()), probs
            categories_probs_all.append(categories_probs)
            categories_all.append(categories)
        categorical_metadata["total_num_categories"] = sum(num_categories_all)
        categorical_metadata["num_categories_all"] = num_categories_all
        categorical_metadata["relative_indices_all"] = np.array(relative_indices_all)
        categorical_metadata["categories_probs_all"] = categories_probs_all
        categorical_metadata["categories_all"] = categories_all

        self.metadata["categorical"] = categorical_metadata

        self.one_hot_enc.fit(x)
        return self.one_hot_enc.transform(x)

    def fit(self, x, **kwargs):
        # we've got nothing to fit in this case
        pass

    def transform(self, x, **kwargs):
        total_transformed_features = 0
        x_numerical, x_categorical = None, None
        if self.num_numerical_features > 0:
            x_numerical = self.transform_numerical_data(x[self.numerical_features])
            self.metadata["numerical"] = self.mode_specific_normalizer.metadata
            total_transformed_features += (self.metadata["numerical"]["relative_indices_all"][-1] +
                                           self.metadata["numerical"]["num_valid_clusters_all"][-1])
        if self.num_categorical_features > 0:
            x_categorical = self.transform_categorical_data(x[self.categorical_features])
            total_transformed_features += (self.metadata["categorical"]["relative_indices_all"][-1] +
                                           self.metadata["categorical"]["num_categories_all"][-1] + 1)

        # since we concatenate the categorical features AFTER the numerical alphas and betas
        # so we'll create an overall relative indices array where we offset the relative indices
        # of the categorical features by the total number of numerical features components
        relative_indices_all = []
        offset = 0
        if x_numerical is not None:
            cont_relative_indices = self.metadata["numerical"]["relative_indices_all"]
            relative_indices_all.extend(cont_relative_indices)
            offset = cont_relative_indices[-1] + self.metadata["numerical"]["num_valid_clusters_all"][-1]
        if x_categorical is not None:
            # +1 since the categorical relative indices start at 0
            relative_indices_all.extend(self.metadata["categorical"]["relative_indices_all"] + 1 + offset)
        self.metadata["relative_indices_all"] = relative_indices_all
        self.metadata["total_transformed_features"] = total_transformed_features
        x_transformed = np.concatenate([x_numerical, x_categorical.toarray()], axis=1)
        return x_transformed

    def reverse_transform(self, x_generated):
        """
        Reverses transforms the generated data to the original data format.

        Args:
            x_generated: Generated dataset.

        Returns:
            Generated data in the original data format.
        """
        all_features = self.numerical_features + self.categorical_features
        if self.num_numerical_features > 0:
            num_valid_clusters_all = self.metadata["numerical"]["num_valid_clusters_all"]
        if self.num_categorical_features > 0:
            num_categories_all = self.metadata["categorical"]["num_categories_all"]
        data = {}
        cat_index = 0       # categorical index
        cont_index = 0      # numerical index
        for index, feature_name in enumerate(all_features):
            # the first n features are numerical
            if index < len(self.numerical_features):
                alphas = x_generated[:, index]
                betas = x_generated[:, index + 1 : index + 1 + num_valid_clusters_all[cont_index]]
                # Recall that betas represent the one hot encoded form of the cluster number
                cluster_indices = np.argmax(betas, axis=1)
                # List of cluster means for a feature. contains one value per cluster
                means = self.metadata["numerical"][feature_name]["clusters_means"]

                # Since each individual element within the cluster is associated with
                # one of the cluster's mean. We use the `cluster_indices` to get
                # a list of size len(x) where each element is a mean for the corresponding
                # element in the feature
                means = means[cluster_indices]
                # Do the same for stds
                stds = self.metadata["numerical"][feature_name]["clusters_stds"]
                stds = stds[cluster_indices]
                feature = alphas * (self.std_multiplier * stds) + means
                data[feature_name] = feature
                cont_index += 1
            else:
                # if the index is greater than or equal to len(numerical_features),
                # then the column at hand is categorical
                raw_feature_parts = x_generated[:, index: index + num_categories_all[cat_index]]
                categories = self.one_hot_enc.categories_[cat_index]
                categories_indices = tf.argmax(raw_feature_parts, axis=1)
                feature = categories[categories_indices]
                data[feature_name] = feature
                cat_index += 1
        return pd.DataFrame(data)


class CTGANDataSampler:
    """
    CTGANDataSampler class based on the data sampler class
    in the official CTGAN implementation.

    Reference(s):
        https://arxiv.org/abs/1907.00503
        https://github.com/sdv-dev/CTGAN/

    Args:
        batch_size: ``int``, default 512,
            Batch size to use for the dataset.

        categorical_features: ``List[str]``,
            List of categorical features names

        numerical_features: ```List[str]``,
            List of numerical features names

        metadata: ``dict``,
            A dictionary of metadata computed during data transformation.
            You can access it from the ``.get_metadata()`` of ``CTGANDataTransformer`` instance.
    """
    def __init__(self,
                 batch_size=512,
                 categorical_features=None,
                 numerical_features=None,
                 metadata=None):
        if categorical_features is None:
            raise ValueError("`categorical_features` must not be None. "
                             "CTGAN requires dataset to have atleast one categorical feature, "
                             "if your dataset doesn't contain any categorical features, "
                             "then you should use some other generative model. ")

        if metadata is None:
            raise ValueError("`metadata` must not be None. "
                             "DataSampler requires meta data computed in the transformation step "
                             "to sample data. "
                             "Use DataTransformer's `get_metadata` method to "
                             "access meta data and pass it as argument to DataSampler.")

        self.batch_size = batch_size
        self.categorical_features = categorical_features
        self.numerical_features = numerical_features
        self.metadata = metadata

        self.num_samples = None
        self.data_dim = None
        self.batch_size = batch_size
        self.row_idx_by_categories = list()

    def get_dataset(self,
                    x_transformed,
                    x_original=None):
        """
        Args:
            x_transformed: Dataset transformed using DataTransformer class
            x_original: Original Dataset - a pandas DataFrame.
                It is used for computing categorical values' probabilities
                for later sampling.
        Returns:
            Returns a tensorflow dataset that utilizes the sample_data method
            to create batches of data. This way user can just pass the dataset object to the fit
            method of the model and each batch generated will satisfy all out requirements of sampling
        """
        self.num_samples, self.data_dim = x_transformed.shape
        # adapting the approach from the official implementation
        # to sample evenly across the categories to combat imbalance
        row_idx_raw = [x_original.groupby(feature).groups for feature in self.categorical_features]
        self.row_idx_by_categories = tf.ragged.constant([[values.to_list()
                                                          for values in feat.values()]
                                                         for feat in row_idx_raw])

        total_num_categories = self.metadata["categorical"]["total_num_categories"]

        dataset = tf.data.Dataset.from_generator(
            self.generator,
            output_signature=(tf.TensorSpec(shape=(self.batch_size, tf.shape(x_transformed)[1]),
                                            name="real_samples"),
                              tf.TensorSpec(shape=(self.batch_size, total_num_categories),
                                            dtype=tf.float32, name="cond_vectors_real"),
                              tf.TensorSpec(shape=(self.batch_size, total_num_categories),
                                            dtype=tf.float32, name="cond_vectors"),
                              tf.TensorSpec(shape=(self.batch_size, len(self.categorical_features)),
                                            dtype=tf.float32, name="mask")),
            args=[x_transformed]
        )
        return dataset

    def sample_cond_vectors_for_training(self,
                                         random_features_idx=None,
                                         random_values_idx=None):
        # 1. Create Nd zero-filled mask vectors mi = [mi(k)] where k=1...|Di| and for i = 1,...,Nd,
        # so the ith mask vector corresponds to the ith column,
        # and each component is associated to the category of that column.
        masks = tf.zeros([self.batch_size, len(self.categorical_features)])
        cond_vectors = tf.zeros([self.batch_size, self.metadata["categorical"]["total_num_categories"]])
        # 2. Randomly select a discrete column Di out of all the Nd discrete columns, with equal probability.
        # >>> We select them in generator method

        random_features_relative_indices = tf.gather(self.metadata["categorical"]["relative_indices_all"],
                                                     indices=random_features_idx)

        # 3. Construct a PMF across the range of values of the column selected in 2, Di* , such that the
        # probability mass of each value is the logarithm of its frequency in that column.
        # >>> Moved to the generator method, which calls this method
        # and passes the value of `random_features_idx`

        # Choose random value index for each feature
        # >>> Moved to the generator method, which calls this method
        # and passes the value of `random_values_idx`

        # Offset this index by relative index of the feature that it belongs to.
        # because the final cond vector is the concatenation of all features and
        # is just one vector that has the length equal to total_num_categories
        random_values_idx_offsetted = random_values_idx + tf.cast(random_features_relative_indices,
                                                                  dtype=tf.int32)
        # Indices are required by the tensor_scatter_nd_update method to be of type int32 and NOT int64
        # random_values_idx_offsetted = tf.cast(random_values_idx_offsetted, dtype=tf.int32)
        # indices_cond = list(zip(tf.range(batch_size), random_values_idx_offsetted))
        indices_cond = tf.stack([tf.range(self.batch_size), random_values_idx_offsetted], axis=1)
        ones = tf.ones(self.batch_size)
        cond_vectors = tf.tensor_scatter_nd_update(cond_vectors, indices_cond, ones)
        # indices_mask = list(zip(tf.range(batch_size), tf.cast(random_features_idx, dtype=tf.int32)))
        indices_mask = tf.stack([tf.range(self.batch_size), random_features_idx], axis=1)
        masks = tf.tensor_scatter_nd_update(masks, indices_mask, tf.ones(self.batch_size))
        return cond_vectors, masks

    def sample_cond_vectors_for_generation(self, batch_size):
        """
        The difference between this method and the training one is that, here
        we sample indices purely randomly instead of based on the calculated
        probability as proposed in the paper.
        """
        num_categories_all = self.metadata["categorical"]["num_categories_all"]
        cond_vectors = tf.zeros((batch_size, self.metadata["categorical"]["total_num_categories"]))
        random_features_idx = tf.random.uniform(shape=(batch_size,),
                                                minval=0, maxval=len(self.categorical_features),
                                                dtype=tf.int32)

        # For each randomly picked feature, we get it's corresponding num_categories
        random_num_categories_all = tf.gather(num_categories_all,
                                              indices=random_features_idx)
        # Then we select one category index from a feature using a range of 0 — num_categories
        random_values_idx = [tf.squeeze(tf.random.uniform(shape=(1,),
                                                          minval=0, maxval=num_categories, dtype=tf.int32))
                             for num_categories in random_num_categories_all]
        random_values_idx = tf.stack(random_values_idx)
        # Offset this index by relative index of the feature that it belongs to.
        # because the final cond vector is the concatenation of all features and
        # is just one vector that has the length equal to total_num_categories
        random_features_relative_indices = tf.gather(self.metadata["categorical"]["relative_indices_all"],
                                                     indices=random_features_idx)
        random_values_idx_offsetted =  random_values_idx + tf.cast(random_features_relative_indices, dtype=tf.int32)
        # random_values_idx_offsetted = tf.cast(random_values_idx_offsetted, dtype=tf.int32)
        indices_cond = list(zip(tf.range(batch_size), random_values_idx_offsetted))
        ones = tf.ones(batch_size)
        cond_vectors = tf.tensor_scatter_nd_update(cond_vectors, indices_cond, ones)
        return cond_vectors

    def generator(self, x_transformed, for_tvae=False):
        """
        Used to create a tensorflow dataset.
        Returns:
            A batch of data
        """
        # This random_feature_indices variable is required during the sample_cond vector method
        # but since we're using sample_data function to create out tensorflow dataset, this
        # gets called first to generate a batch, so keep in mind that this is where this
        # variable gets its values. We could alternatively just return these indices
        # and pass them as argument to the sample cond_vec but for now let's just work with it.
        num_steps_per_epoch = self.num_samples // self.batch_size
        for _ in range(num_steps_per_epoch):
            random_features_idx = tf.random.uniform([self.batch_size], minval=0,
                                                    maxval=len(self.categorical_features),
                                                    dtype=tf.int32)
            # NOTE: We've precomputed the probabilities in the DataTransformer class for each feature already
            # to speed things up.
            random_features_categories_probs = tf.gather(
                tf.ragged.constant(self.metadata["categorical"]["categories_probs_all"]),
                indices=random_features_idx)
            random_values_idx = [tf_random_choice(np.arange(len(feature_probs)),
                                                  n_samples=1,
                                                  p=feature_probs)
                                 for feature_probs in random_features_categories_probs]
            random_values_idx = tf.cast(tf.squeeze(random_values_idx), dtype=tf.int32)
            # the official implementation uses actual indices during the sample_cond_vector method
            # but uses the shuffled version in sampling data, so we're gonna do just that.
            shuffled_idx = tf.random.shuffle(tf.range(self.batch_size))
            # features_idx = random_features_idx
            # values_idx = random_values_idx
            shuffled_features_idx = tf.gather(random_features_idx, indices=shuffled_idx)
            shuffled_values_idx = tf.gather(random_values_idx, indices=shuffled_idx)
            sample_idx = []
            # TODO make it more efficient -- maybe replace with gather or gathernd or something
            for feat_id, val_id in zip(shuffled_features_idx, shuffled_values_idx):
                s_id = tf_random_choice(self.row_idx_by_categories[tf.squeeze(feat_id)][tf.squeeze(val_id)],
                                        n_samples=1)
                sample_idx.append(tf.squeeze(s_id))

            # we also return shuffled_idx because it will be required to shuffle the conditional vector
            # in the training loop as we want to keep the shuffling consistent as the batch of
            # transformed data and cond vector must have one to one feature correspondence.
            # yield x_transformed[sample_idx], shuffled_idx, random_features_idx, random_values_idx

            # `cond_vectors` will be first concatenated with the noise
            # vector `z` to create generator input and then will be concatenated
            # with the generated samples to serve as input for discriminator
            cond_vectors, mask = self.sample_cond_vectors_for_training(random_features_idx=random_features_idx,
                                                                       random_values_idx=random_values_idx)
            # `cond_vectors_real` will be concatenated with the real_samples
            # and passed to the discriminator
            cond_vectors_real = tf.gather(cond_vectors, indices=shuffled_idx)
            real_samples = x_transformed[sample_idx]
            yield real_samples, cond_vectors_real, cond_vectors, mask



import tensorflow as tf
from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class TabTransformer(keras.Model):
    """
    TabTransformer model class with LayerFlow design.

    TabTransformer architecture is proposed by Xin Huang et al.
    in the paper,
    TabTransformer: Tabular Data Modeling Using Contextual Embeddings.

    TabTransformer is a novel deep tabular data modeling architecture for
    supervised and semi-supervised learning.
    The TabTransformer is built upon self-attention based Transformers.
    The Transformer layers transform the embeddings of categorical features
    into robust contextual embeddings to achieve higher prediction accuracy.

    Reference(s):
        https://arxiv.org/abs/2012.06678

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset.

        categorical_feature_embedding: ``layers.Layer``,
            An instance of ``CategoricalFeatureEmbedding`` layer to embedd categorical features
            or any layer that can work in its place for that purpose.
            If None, a ``CategoricalFeatureEmbedding`` layer with default values will be used.
            You can import the ``CategoricalFeatureEmbedding`` layer as follows,
                >>> from teras.layerflow.layers import CategoricalFeatureEmbedding

        column_embedding: ``layers.Layer``,
            An instance of ``TabTColumnEmbedding`` layer to apply over categorical embeddings,
            or any layer that can work in its place for that purpose.
            If None, a ``TabTColumnEmbedding`` layer with default values will be used.
            You can import the ``TabTColumnEmbedding`` layer as follows,
                >>> from teras.layers import TabTransformerColumnEmbedding

        encoder: ``layers.Layer``,
            An instance of ``Encoder`` layer to encode feature embeddings,
            or any layer that can work in its place for that purpose.
            If None, an ``Encoder`` layer with default values will be used.
            You can import the ``Encoder`` layer as follows,
                >>> from teras.layerflow.layers import Encoder

        numerical_feature_normalization: ``layers.Layer``,
            An instance of ``NumericalFeatureNormalization`` layer to normalize numerical features,
            or any layer that can work in its place for that purpose.
            If None, an ``NumericalFeatureNormalization`` layer with default values will be used.
            You can import the ``NumericalFeatureNormalization`` layer as follows,
                >>> from teras.layerflow.layers import NumericalFeatureNormalization

        head: ``keras.layers.Layer``,
            An instance of either ``ClassificationHead`` or ``RegressionHead`` layers,
            depending on the task at hand.

            REMEMBER: In case you're using this model as a base model for pretraining, you MUST leave
            this argument as None.

            You can import the ``ClassificationHead`` and ``RegressionHead`` layers as follows,
                >>> from teras.layers import ClassificationHead
                >>> from teras.layers import RegressionHead
    """
    def __init__(self,
                 input_dim: int,
                 categorical_feature_embedding: keras.layers.Layer = None,
                 column_embedding: keras.layers.Layer = None,
                 encoder: keras.layers.Layer = None,
                 numerical_feature_normalization: keras.layers.Layer = None,
                 head: keras.layers.Layer = None,
                 **kwargs):
        if categorical_feature_embedding is None and numerical_feature_normalization is None:
            raise ValueError("Both `categorical_feature_embedding` and `numerical_feature_normalization` "
                             "cannot be None at the same time as a tabular dataset must contains "
                             "features of at least one of the types if not both. ")

        if isinstance(input_dim, int):
            input_dim = (input_dim,)
        inputs = keras.layers.Input(shape=input_dim,
                                    name="inputs")
        categorical_out = None
        if categorical_feature_embedding is not None:
            x = categorical_feature_embedding(inputs)

            if column_embedding is not None:
                x = column_embedding(x)

            if encoder is None:
                raise ValueError("`encoder` is required to encode the categorical embedding, "
                                 "but received `None`. "
                                 "Please pass an instance of `Encode` layer. "
                                 "You can import it as, `from teras.layerflow.layers import Encoder`")
            x = encoder(x)
            x = keras.layers.Flatten()(x)
            categorical_out = x

        if numerical_feature_normalization is not None:
            numerical_out = numerical_feature_normalization(inputs)
            if categorical_out is not None:
                x = keras.layers.Concatenate()([categorical_out, numerical_out])
            else:
                x = numerical_out

        outputs = x
        if head is not None:
            outputs = head(outputs)

        super().__init__(inputs=inputs, outputs=outputs, **kwargs)

        self.input_dim = input_dim
        self.categorical_feature_embedding = categorical_feature_embedding
        self.column_embedding = column_embedding
        self.encoder = encoder
        self.numerical_feature_normalization = numerical_feature_normalization
        self.head = head

    def get_config(self):
        config = super().get_config()
        config.update({'input_dim': self.input_dim,
                       'categorical_feature_embedding': keras.layers.serialize(self.categorical_feature_embedding),
                       'column_embedding': keras.layers.serialize(self.column_embedding),
                       'encoder': keras.layers.serialize(self.encoder),
                       'numerical_feature_normalization': keras.layers.serialize(self.numerical_feature_normalization),
                       'head': keras.layers.serialize(self.head),
                       })
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        categorical_feature_embedding = keras.layers.deserialize(config.pop("categorical_feature_embedding"))
        column_embedding = keras.layers.deserialize(config.pop("column_embedding"))
        encoder = keras.layers.deserialize(config.pop("encoder"))
        numerical_feature_normalization = keras.layers.deserialize(config.pop("numerical_feature_normalization"))
        head = keras.layers.deserialize(config.pop("head"))
        return cls(input_dim=input_dim,
                   categorical_feature_embedding=categorical_feature_embedding,
                   column_embedding=column_embedding,
                   encoder=encoder,
                   numerical_feature_normalization=numerical_feature_normalization,
                   head=head,
                   **config)


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class TabTransformerPretrainer(keras.Model):
    """
    Pretrainer model for TabTransformer based on the
    Replaced Token Detection (RTD) method.
    RTD replaces the original feature by a random value
    of that feature.
    Here, the loss is minimized for a binary classifier
    that tries to predict whether or not the feature has
    been replaced.

    Reference(s):
        https://arxiv.org/abs/2012.06678

    Args:
        model: ``TabTransformer``,
            An instance of base ``TabTransformer`` class to pretrain.

        features_metadata: ``dict``,
            a nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        replace_rate: ``float``, default 0.3,
            Fraction of total features per sample to replace.
            Must be in between 0. - 1.0
    """

    def __init__(self,
                 model: TabTransformer,
                 features_metadata: dict,
                 replace_rate: float = 0.3,
                 **kwargs):
        super().__init__(**kwargs)
        self.model = model
        self.replace_rate = replace_rate
        self.features_metadata = features_metadata

        self._categorical_features_exist = len(self.features_metadata["categorical"]) > 0
        self.num_features = len(self.features_metadata["categorical"]) + len(self.features_metadata["numerical"])
        self.num_features_to_replace = tf.cast(tf.cast(self.num_features, tf.float32) * self.replace_rate,
                                               dtype=tf.int32)
        self.loss_tracker = keras.metrics.Mean(name="loss")
        self.head = keras.layers.Dense(self.num_features,
                                       activation="sigmoid",
                                       name="pretrainer_head")

    def get_pretrained_model(self):
        """Returns pretrained model"""
        return self.model

    @property
    def pretrained_model(self):
        """Returns pretrained model"""
        return self.get_pretrained_model()

    def compile(self,
                loss=keras.losses.BinaryCrossentropy(),
                optimizer=keras.optimizers.AdamW(learning_rate=0.01),
                **kwargs):
        super().compile(**kwargs)
        self.loss_fn = loss
        self.optimizer = optimizer

    def call(self, inputs, mask=None):
        # Since in RTD, for a sample, we randomly replace k% of its features values using
        # random values of those features.
        # We can efficiently achieve this by first getting x_rand = shuffle(inputs)
        # then, to apply replacement, inputs = (inputs * (1-mask)) + (x_rand * mask)
        x_rand = tf.random.shuffle(inputs)
        inputs = (inputs * (1 - mask)) + (x_rand * mask)
        intermediate_features = self.model(inputs)
        # Using sigmoid we essentially get a predicted mask which can we used to
        # compute loss just like that of binary classification
        mask_pred = self.head(intermediate_features)
        return mask_pred

    def train_step(self, data):
        if isinstance(data, tuple):
            data = data[0]

        # Feature indices is of the shape batch_size x num_features_to_replace
        feature_indices_to_replace = tf.random.uniform((tf.shape(data)[0], self.num_features_to_replace),
                                                       maxval=self.num_features,
                                                       dtype=tf.int32)
        # Mask is of the shape batch_size x num_features and contains values in range [0, 1]
        # A value of 1 represents if the feature is replaced, while 0 means it's not replaced.
        mask = tf.reduce_max(tf.one_hot(feature_indices_to_replace,
                                        depth=self.num_features),
                             axis=1)

        with tf.GradientTape() as tape:
            mask_pred = self(data, mask=mask)
            loss = self.loss_fn(mask, mask_pred)
        gradients = tape.gradient(loss, self.trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_weights))
        self.loss_tracker.update_state(loss)
        self.compiled_metrics.update_state(mask, mask_pred)
        results = {m.name: m.result() for m in self.metrics}
        return results

    def get_config(self):
        config = super().get_config()
        new_config = {'model': keras.layers.serialize(self.model),
                      'features_metadata': self.features_metadata,
                      'replace_rate': self.replace_rate
                      }
        config.update(new_config)
        return config

    @classmethod
    def from_config(cls, config):
        model = keras.layers.deserialize(config.pop("model"))
        features_metadata = config.pop("features_metadata")
        return cls(model=model,
                   features_metadata=features_metadata,
                   **config)



from tensorflow import keras
from teras.utils.types import ActivationType


@keras.saving.register_keras_serializable(package="teras.layers.vime")
class VimePredictor(keras.layers.Layer):
    """
    Predictor layer based on the architecture proposed by Jinsung Yoon et a.
    in the paper "VIME: Extending the Success of Self- and
    Semi-supervised Learning to Tabular Domain"

    Reference(s):
        https://proceedings.neurips.cc/paper/2020/hash/7d97667a3e056acab9aaf653807b4a03-Abstract.html

    Args:
        num_labels: `int`, default 32,
            Number of labels to predict

        hidden_dim: ``int``, default 32,
            The hidden dimensionality of the predictor.

        activation: ``str`` or ``callable`` or ``keras.layers.Layer``, default "relu",
            Activation function to use in for the hidden layers.
    """
    def __init__(self,
                 num_labels: int = 2,
                 hidden_dim: int = 32,
                 activation: ActivationType = "relu",
                 **kwargs):
        super().__init__(**kwargs)
        self.num_labels = num_labels
        self.hidden_dim = hidden_dim
        self.activation = activation

        self.predictor_block = keras.models.Sequential(name="predictor_block")
        self.inter_layer_1 = keras.layers.Dense(self.hidden_dim,
                                                activation=self.activation,
                                                name="inter_layer_1")
        self.predictor_block.add(self.inter_layer_1)
        self.inter_layer_2 = keras.layers.Dense(self.hidden_dim,
                                                activation=self.activation,
                                                name="inter_layer_2")
        self.predictor_block.add(self.inter_layer_2)
        self.dense_out = keras.layers.Dense(self.num_labels,
                                            activation=None,
                                            name="dense_out")
        self.predictor_block.add(self.dense_out)
        self.softmax = keras.layers.Softmax()

    def call(self, inputs):
        y_hat_logit = self.predictor_block(inputs)
        y_hat = self.softmax(y_hat_logit)
        return y_hat_logit, y_hat

    def get_config(self):
        config = super().get_config()
        activation_serialized = self.activation
        if not isinstance(self.activation, str):
            activation_serialized = keras.layers.serialize(self.activation)
        config.update({'num_labels': self.num_labels,
                       'hidden_dim': self.hidden_dim,
                       'activation': activation_serialized,
                       }
                      )
        return config



import tensorflow as tf
from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.ft_transformer")
class FTCLSToken(keras.layers.Layer):
    """
    CLS Token as proposed and implemented by Yury Gorishniy et al.
    in the paper Revisiting Deep Learning Models for Tabular Data
    in their FTTransformer architecture.

    It adds a BERT-like CLS token to the inputs, which is then used
    for the downstream task such as classification or regression.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        embedding_dim: ``int``, default 32,
            Embedding dimensions used to embed the numerical
            and categorical features.

        initialization: ``str``, default "normal",
            Initialization method to use for the weights.
    """
    def __init__(self,
                 embedding_dim: int = 32,
                 initialization: str = "normal",
                 **kwargs):
        super().__init__(**kwargs)
        self.embedding_dim = embedding_dim
        self.initialization = initialization
        self.token = None

    def build(self, input_shape):
        initializer = self.initialization
        if isinstance(initializer, str):
            initializer = keras.initializers.get(initializer)
        self.token = self.add_weight(initializer=initializer,
                                     shape=(1, self.embedding_dim))

    def call(self, inputs):
        # We want to append the CLS token to the inputs
        # The expected input shape is, (batch_size, num_features, embedding_dim)
        # The self.token has the shape, (1, embedding_dim),
        # since batch_size can be dynamic, in that the last batch may have different shape
        # so we broadcast `self.token` to (`batch_size`, 1, `embedding_dim`), which in simple words
        # copies the `self.token` `batch_size` times along the first dimension.
        # And then we concatenate the `token_broadcasted` with `inputs` along the SECOND dimension,
        # i.e. the feature dimension, which gives us the final inputs of shape
        # (`batch_size`, `num_features + 1`, `embedding_dim`),
        # mind the plus one to `num_features`!
        token_broadcasted = tf.broadcast_to(self.token, shape=(tf.shape(inputs)[0], 1, self.embedding_dim))
        return tf.concat([inputs, token_broadcasted], axis=1)

    def get_config(self):
        config = super().get_config()
        config.update({'embedding_dim': self.embedding_dim,
                       'initialization': self.initialization})
        return config



import tensorflow as tf
from teras.layers.dnfnet.dnnf import DNNF


def test_dnnf_valid_call():
    dnnf = DNNF(num_formulas=16)
    inputs = tf.ones((10, 8), dtype=tf.float32)
    outputs = dnnf(inputs)


def test_dnnf_output_shape():
    dnnf = DNNF(num_formulas=16)
    inputs = tf.ones((8, 10), dtype=tf.float32)
    outputs = dnnf(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 16



from tensorflow import keras
from teras.activations import sparsemax


@keras.saving.register_keras_serializable("teras.layers.tabnet")
class TabNetAttentiveTransformer(keras.layers.Layer):
    """
    Attentive Transformer layer for mask generation
    as proposed by Sercan et al. in TabNet paper.
    It applies a ``Dense`` layer followed by a ``BatchNormalization`` layer
    followed by a ``sparsemax`` activation function.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        data_dim: ``int``,
            Number of features in the input dataset.

        batch_momentum: ``float``, default 0.9,
            Momentum value to use for BatchNormalization layer.

        virtual_batch_size: ``int``, default 64,
            Batch size to use for ``virtual_batch_size`` parameter in ``BatchNormalization`` layer.

        relaxation_factor: ``float``, default 1.5,
            Relaxation factor that promotes the reuse of each feature at different
            decision steps. When it is 1, a feature is enforced to be used only at one
            decision step and as it increases, more flexibility is provided to use a
            feature at multiple decision steps.
            An optimal value of relaxation_factor can have a major role on the performance.
            Typically, a larger value for ``num_decision_steps`` favors for a larger ``relaxation_factor``.
    """
    def __init__(self,
                 data_dim: int,
                 batch_momentum: float = 0.9,
                 virtual_batch_size: int = 64,
                 relaxation_factor: float = 1.5,
                 **kwargs):
        super().__init__(**kwargs)
        # We can't infer data dimensionality from build method because
        # the attentive transformer is applied after feature transformer
        # and hence received hidden representations of arbitrary dimensions
        self.data_dim = data_dim
        self.batch_momentum = batch_momentum
        self.virtual_batch_size = virtual_batch_size
        self.relaxation_factor = relaxation_factor
        self.dense = keras.layers.Dense(self.data_dim, use_bias=False)
        self.batch_norm = keras.layers.BatchNormalization(momentum=batch_momentum,
                                                          virtual_batch_size=self.virtual_batch_size)

    def call(self, inputs, prior_scales=None):
        outputs = self.dense(inputs)
        outputs = self.batch_norm(outputs)
        outputs *= prior_scales
        outputs = sparsemax(outputs)
        return outputs

    def get_config(self):
        config = super().get_config()
        config.update({'data_dim': self.data_dim,
                       'batch_momentum': self.batch_momentum,
                       'virtual_batch_size': self.virtual_batch_size,
                       'relaxation_factor': self.relaxation_factor,
                       })
        return config

    @classmethod
    def from_config(cls, config):
        data_dim = config.pop("data_dim")
        return cls(data_dim=data_dim, **config)



import tensorflow as tf
import numpy as np
import pandas as pd
from teras.preprocessing.gain import (GAINDataTransformer as _GAINDataTransformer,
                                      GAINDataSampler as _GAINDataSampler)
from teras.utils.types import FeaturesNamesType


class PCGAINDataTransformer(_GAINDataTransformer):
    """
    PCGAINDataTransformer class that performs the required transformations
    on the raw dataset required by the PCGAIN architecture.

    It is simply a wrapper around the GAINDataTransformer class, as everything
    is exactly the same.

    Args:
        categorical_features: ``List[str]``,
            List of categorical features names in the dataset.
            Categorical features are encoded by ordinal encoder method.
            And then MinMax normalization is applied.

        numerical_features: ``List[str]``,
            List of numerical features names in the dataset.
            Numerical features are encoded using MinMax normalization.
    """
    def __init__(self,
                 categorical_features: FeaturesNamesType = None,
                 numerical_features: FeaturesNamesType = None
                 ):
        super().__init__(categorical_features=categorical_features,
                         numerical_features=numerical_features)


class PCGAINDataSampler(_GAINDataSampler):
    """
    DataSampler class for PC-GAIN.
    It extends GAIN's DataSampler class because PC-GAIN
    in itself is an extension of GAIN architecture,
    so the data required by them is in similar format.
    But PC-GAIN requires one extra thing -- the pretraining dataset.
    Hence, the need to extend the GAIN Data Sampler class and
    implement that functionality here.

    Reference(s):
        https://arxiv.org/abs/2011.07770

    Args:
        batch_size: ``int``, default 512,
            Batch size to use for dataset.

        shuffle: ``bool``, default True,
            Whether to shuffle the dataset.

        random_seed: ``int``, default None,
            Random seed to make shuffling
            and any other random operations consistent.
    """
    def __init__(self,
                 batch_size=512,
                 shuffle=True,
                 random_seed=None):
        super().__init__(batch_size=batch_size,
                         shuffle=shuffle,
                         random_seed=random_seed)

    def get_pretraining_dataset(self,
                                x_transformed,
                                pretraining_size: float = 0.2,
                                batch_size: int = None):
        """
        Args:
            x_transformed:
                Transformed dataset.

            pretraining_size: ``float``, default 0.2,
                Fraction of dataset that will be used for pretraining.
                As a general rule of thumb (based on official implementation),
                    - for datasets with ``num_samples >= 20000``, use ``pretraining_size=0.2``
                    - for datasets with ``num_samples < 20000``, use ``pretraining_size = 0.4``
                Please note that, this is NOT a hard and fast rule. You can specify
                whatever fraction size you deem reasonable for your use case.

            batch_size: ``int``,
                In case you want to use a different batch size for your
                pretraining than for actual training. If None, the ``batch_size``
                specified while instantiating ``PCGAINDataSampler`` will be used.
        """
        if batch_size is None:
            batch_size = self.batch_size

        # Sort all the samples according to their missing rate and select the top
        # lambda a.k.a. `pretraining_size` samples to make the pretraining dataset.
        # Missing rate in simple terms means,
        # the number of nans per row divided by the total number of element in the row.
        # Or in other terms, number of nans per row divided by the dimensionality of dataset.
        if isinstance(x_transformed, pd.DataFrame):
            x_transformed = x_transformed.values
        num_samples, dim = x_transformed.shape
        num_pretraining_samples = int(pretraining_size * num_samples)
        missing_rate_per_row = np.sum(np.isnan(x_transformed), axis=1) / dim
        # Note that the below indices will be in ascending order
        sorted_idx = np.argsort(missing_rate_per_row)
        # Since sorted idx are in ascending order and we need `pretraining_size` fraction
        # of samples with least missing rate. so we select the first `num_pretraining_samples` idx
        pretraining_samples_idx = sorted_idx[:num_pretraining_samples]

        # x_transformed but with only pretraining samples
        x_transformed = x_transformed[pretraining_samples_idx]

        # We need separate batches of data for generator and discriminator on each step as required by
        # GAIN and hence PC-GAIN
        gen_idx = np.random.permutation(np.arange(len(x_transformed)))
        disc_idx = np.random.permutation(np.arange(len(x_transformed)))

        pretraining_dataset = tf.data.Dataset.from_generator(
            generator=self._pretraining_dataset_generator,
            output_signature=(tf.TensorSpec(shape=(None, dim), dtype=tf.float32, name="generator_pretraining_data"),
                              tf.TensorSpec(shape=(None, dim), dtype=tf.float32, name="discriminator_pretraining_data")
                              ),
            args=[x_transformed, gen_idx, disc_idx, batch_size]
        )
        return pretraining_dataset

    def _pretraining_dataset_generator(self, x, gen_idx, disc_idx, batch_size):
        """
         Generator function that is used to create pretraining tesnorflow dataset.
         It applies any manipulations that are required and then generates
         batches in the format required by the architecture.
         """
        num_samples, dim = x.shape
        steps_per_epoch = num_samples // self.batch_size
        is_n_divisible_by_batch_size = num_samples % self.batch_size == 0
        steps_per_epoch += 1 if not is_n_divisible_by_batch_size else 0
        from_index = 0
        to_index = batch_size
        for i in range(steps_per_epoch):
            # On the last batch when the number of samples isn't divisible
            # by batch size, then trying to extract x[from_index: to_index] will
            # result in error since to_index will go out of bounds.
            # So to avoid this, we add the following check
            # and modify the to_index value accordingly.
            if i == (steps_per_epoch - 1) and not is_n_divisible_by_batch_size:
                x_generator = x[gen_idx[from_index: None]]
                x_discriminator = x[disc_idx[from_index: None]]
            else:
                x_generator = x[gen_idx[from_index: to_index]]
                x_discriminator = x[disc_idx[from_index: to_index]]
                from_index += batch_size
                to_index += batch_size

            yield x_generator, x_discriminator



from tensorflow import keras
from teras.models.gain import (GAINGenerator as _GAINGenerator,
                               GAINDiscriminator as _GAINDiscriminator,
                               GAIN)
from teras.utils.types import UnitsValuesType, ActivationType
from teras.layerflow.models.pcgain import PCGAIN as _PCGAIN_LF


@keras.saving.register_keras_serializable(package="teras.models")
class PCGAINGenerator(_GAINGenerator):
    """
    Generator model for the PCGAIN architecture.
    It is exact as the ``GAINGenerator`` model, in fact, this class is just a wrapper
    around the ``GAINGenerator`` model class.

    Args:
        data_dim: ``int``,
            The dimensionality of the input dataset.

            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

            You can access the dimensionality of the transformed dataset through the
            ``.data_dim`` attribute of the ``PCGAINDataSampler`` instance used in sampling
            the dataset.

        units_values: ``List[int]`` or ``Tuple[int]``,
            A list or tuple of units for constructing hidden block.
            For each value, a ``GAINGeneratorBlock``
            (``from teras.layers import GAINGeneratorBlock``)
            of that dimensionality (units) is added to the generator
            to form the ``hidden block`` of the generator.
            By default, ``units_values`` = [``data_dim``, ``data_dim``].

        activation_hidden: default "relu",
            Activation function to use for the hidden layers in the hidden block.

        activation_out: default "sigmoid",
            Activation function to use for the output layer of the Generator.
        ```
    """
    def __init__(self,
                 data_dim: int,
                 units_values: UnitsValuesType = None,
                 activation_hidden: ActivationType = "relu",
                 activation_out: ActivationType = "sigmoid",
                 **kwargs):
        super().__init__(data_dim=data_dim,
                         units_values=units_values,
                         activation_hidden=activation_hidden,
                         activation_out=activation_out,
                         **kwargs)


@keras.saving.register_keras_serializable(package="teras.models")
class PCGAINDiscriminator(_GAINDiscriminator):
    """
    Discriminator model for the PCGAIN architecture.
    It is exact as the ``GAINDiscriminator`` model, in fact, this class is just a wrapper
    around the ``GAINDiscriminator`` model class.

    Args:
        data_dim: ``int``,
            The dimensionality of the input dataset.

            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

            You can access the dimensionality of the transformed dataset through the
            ``.data_dim`` attribute of the ``GAINDataSampler`` instance used in sampling
            the dataset.

        units_values: ``List[int]`` or ``Tuple[int]``,
            A list or tuple of units for constructing hidden block.
            For each value, a ``GAINDiscriminatorBlock``
            (``from teras.layers import GAINDiscriminatorBlock``)
            of that dimensionality (units) is added to the discriminator
            to form the ``hidden block`` of the discriminator.
            By default, ``units_values`` = [``data_dim``, ``data_dim``].

        activation_hidden: default "relu",
            Activation function to use for the hidden layers in the hidden block.

        activation_out: default "sigmoid",
            Activation function to use for the output layer of the Discriminator.
        ```
    """
    def __init__(self,
                 data_dim: int,
                 units_values: UnitsValuesType = None,
                 activation_hidden: ActivationType = "relu",
                 activation_out: ActivationType = "sigmoid",
                 **kwargs):
        super().__init__(data_dim=data_dim,
                         units_values=units_values,
                         activation_hidden=activation_hidden,
                         activation_out=activation_out,
                         **kwargs)


@keras.saving.register_keras_serializable(package="teras.models")
class PCGAINClassifier(keras.Model):
    """
    The auxiliary classifier for the PC-GAIN architecture
    proposed by Yufeng Wang et al. in the paper
    "PC-GAIN: Pseudo-label Conditional Generative Adversarial
    Imputation Networks for Incomplete Data"

    Reference(s):
        https://arxiv.org/abs/2011.07770

    Args:
        data_dim: ``int``,
            The dimensionality of the input dataset.
            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

        num_classes: ``int``,
            Number of classes to predict.
            It should be equal to the `num_clusters`,
            computed during the pseudo label generation.

        units_values: ``List[int]`` or ``Tuple[int]``,
            A list/tuple of units to construct hidden block of classifier.
            For each element, a new hidden layer will be added.
            By default, `units_values` = [`data_dim`, `data_dim`]

        activation_hidden: default "relu",
            Activation function to use for the hidden layers in the classifier.

        activation_out: default "softmax",
            Activation function to use for the output layer of classifier.
    """
    def __init__(self,
                 data_dim: int,
                 num_classes: int = None,
                 units_values: UnitsValuesType = None,
                 activation_hidden: ActivationType = "relu",
                 activation_out: ActivationType = "softmax",
                 **kwargs):
        super().__init__(**kwargs)
        self.num_classes = num_classes
        self.units_values = units_values
        self.activation_hidden = activation_hidden
        self.activation_out = activation_out

        if self.units_values is None:
            self.units_values = [data_dim] * 2

        self.hidden_block = keras.models.Sequential(name="classifier_hidden_block")
        for units in self.units_values:
            self.hidden_block.add(keras.layers.Dense(units,
                                                     activation=self.activation_hidden))

        self.output_layer = keras.layers.Dense(self.num_classes,
                                               activation=self.activation_out)

    def call(self, inputs):
        x = self.hidden_block(inputs)
        return self.output_layer(x)

    def get_config(self):
        config = super().get_config()
        new_config = {'data_dim': self.data_dim,
                      'num_classes': self.num_classes,
                      'units_values': self.units_values,
                      'activation_hidden': self.activation_hidden,
                      'activation_out': self.activation_out
                      }
        config.update(new_config)
        return config

    @classmethod
    def from_config(cls, config):
        data_dim = config.pop("data_dim")
        return cls(data_dim=data_dim,
                   **config)


@keras.saving.register_keras_serializable(package="teras.models")
class PCGAIN(_PCGAIN_LF):
    """
    PC-GAIN is a missing data imputation model based
    on the GAIN architecture.

    This implementation is based on the architecture
    proposed by Yufeng Wang et al. in the paper
    "PC-GAIN: Pseudo-label Conditional Generative Adversarial
    Imputation Networks for Incomplete Data"

    Reference(s):
        https://arxiv.org/abs/2011.07770

    Args:
        input_dim: ``int``,
            The dimensionality of the input dataset.

            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

            You can access the dimensionality of the transformed dataset through the
            ``.data_dim`` attribute of the ``PCGAINDataSampler`` instance used in sampling
            the dataset.

        generator_units_values: ``List[int]`` or ``Tuple[int]``,
            A list or tuple of units for constructing hidden block for the Generator.
            For each value, a ``PCGAINDiscriminatorBlock``
            (``from teras.layers import PCGAINGeneratorBlock``)
            of that dimensionality (units) is added to the generator to form the
            `hidden block` of the generator.
            By default, ``generator_units_values = [data_dim, data_dim]``.

        discriminator_units_values: ``List[int]`` or ``Tuple[int]``,
            A list or tuple of units for constructing hidden block for the Discriminator.
            For each value, a ``PCGAINDiscriminatorBlock``
            (``from teras.layers import PCGAINDiscriminatorBlock``)
            of that dimensionality (units) is added to the discriminator
            to form the `hidden block` of the discriminator.
            By default, ``discriminator_units_values = [data_dim, data_dim]``.

        generator_activation_hidden: default "relu",
            Activation function to use for the hidden layers in the hidden block
            for the Generator.

        discriminator_activation_hidden: default "relu",
            Activation function to use for the hidden layers in the hidden block
            for the Discriminator.

        generator_activation_out: default "sigmoid",
            Activation function to use for the output layer of the Generator.

        discriminator_activation_out: default "sigmoid",
            Activation function to use for the output layer of the Discriminator.

        num_discriminator_steps: ``int``, default 1,
            Number of discriminator training steps per PCGAIN training step.

        hint_rate: ``float``, default 0.9,
            Hint rate will be used to sample binary vectors for
            `hint vectors` generation. Should be between 0. and 1.
            Hint vectors ensure that generated samples follow the
            underlying data distribution.

        alpha: ``float``, default 200.,
            Hyper parameter for the generator loss computation that
            controls how much weight should be given to the MSE loss.
            Precisely, ``generator_loss = cross_entropy_loss + alpha * mse_loss``
            The higher the ``alpha``, the more the ``mse_loss`` will affect the
            overall generator loss.

        beta: ``float``, default 100.,
            Hyper parameter for generator loss computation that
            controls the contribution of the classifier's loss to the
            overall generator loss.

        num_clusters: ``int``, default 5,
            Number of clusters to cluster the imputed data
            that is generated during pretraining.
            These clusters serve as pseudo labels for training of classifier.

        clustering_method: ``str``, default "kmeans",
            Should be one of the following,
            ["Agglomerative", "KMeans", "MiniBatchKMeans", "Spectral", "SpectralBiclustering"]
            The names are case in-sensitive.
    """
    def __init__(self,
                 input_dim: int,
                 generator_units_values: UnitsValuesType = None,
                 discriminator_units_values: UnitsValuesType = None,
                 generator_activation_hidden: ActivationType = "relu",
                 discriminator_activation_hidden: ActivationType = "relu",
                 generator_activation_out: ActivationType = "sigmoid",
                 discriminator_activation_out: ActivationType = "sigmoid",
                 num_discriminator_steps: int = 1,
                 hint_rate: float = 0.9,
                 alpha: float = 200.,
                 beta: float = 100.,
                 num_clusters: int = 5,
                 clustering_method: str = "kmeans",
                 **kwargs):
        # Since we pretrain using the EXACT SAME architecture as GAIN
        # so here we simply use the GAIN model, which acts as `pretrainer`.
        # And since it instantiates generators and discriminator models,
        # we can pass these generator and discriminator parameters to GAIN
        # which acts as a proxy pretrainer and have it instantiate and pretrain them,
        # and then we can access those pretrained Generator and Discriminator models
        # and use them here in our PC-GAIN architecture.
        pretrainer = GAIN(input_dim=input_dim,
                          generator_units_values=generator_units_values,
                          generator_activation_hidden=generator_activation_hidden,
                          generator_activation_out=generator_activation_out,
                          discriminator_units_values=discriminator_units_values,
                          discriminator_activation_hidden=discriminator_activation_hidden,
                          discriminator_activation_out=discriminator_activation_out,
                          hint_rate=hint_rate,
                          alpha=alpha
                          )
        classifier = PCGAINClassifier(data_dim=input_dim,
                                      num_classes=num_clusters)

        super().__init__(pretrainer=pretrainer,
                         classifier=classifier,
                         num_discriminator_steps=num_discriminator_steps,
                         hint_rate=hint_rate,
                         alpha=alpha,
                         beta=beta,
                         num_clusters=num_clusters,
                         clustering_method=clustering_method,
                         **kwargs)
        self.input_dim = input_dim
        self.generator_units_values = generator_units_values
        self.discriminator_units_values = discriminator_units_values
        self.generator_activation_hidden = generator_activation_hidden
        self.discriminator_activation_hidden = discriminator_activation_hidden
        self.generator_activation_out = generator_activation_out
        self.discriminator_activation_out = discriminator_activation_out
        self.num_discriminator_steps = num_discriminator_steps
        self.hint_rate = hint_rate
        self.alpha = alpha
        self.beta = beta
        self.num_clusters = num_clusters
        self.clustering_method = clustering_method

    def get_config(self):
        gen_act_hidden_serialized = self.generator_activation_hidden
        if not isinstance(gen_act_hidden_serialized, str):
            gen_act_hidden_serialized = keras.layers.serialize(gen_act_hidden_serialized)

        gen_act_out_serialized = self.generator_activation_out
        if not isinstance(gen_act_out_serialized, str):
            gen_act_out_serialized = keras.layers.serialize(gen_act_out_serialized)

        disc_act_hidden_serialized = self.discriminator_activation_hidden
        if not isinstance(disc_act_hidden_serialized, str):
            disc_act_hidden_serialized = keras.layers.serialize(disc_act_hidden_serialized)

        disc_act_out_serialized = self.discriminator_activation_out
        if not isinstance(disc_act_out_serialized, str):
            disc_act_out_serialized = keras.layers.serialize(disc_act_out_serialized)

        config = {"name": self.name,
                  "trainable": self.trainable,
                  "input_dim": self.input_dim,
                  "generator_units_values": self.generator_units_values,
                  "discriminator_units_values": self.discriminator_units_values,
                  "generator_activation_hidden": gen_act_hidden_serialized,
                  "discriminator_activation_hidden": disc_act_hidden_serialized,
                  "generator_activation_out": gen_act_out_serialized,
                  "discriminator_activation_out": disc_act_out_serialized,
                  "num_discriminator_steps": self.num_discriminator_steps,
                  "hint_rate": self.hint_rate,
                  "alpha": self.alpha,
                  "beta": self.beta,
                  "num_clusters": self.num_clusters,
                  "clustering_method": self.clustering_method,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        return cls(input_dim=input_dim,
                   **config)



import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend as K
from teras.losses.tvae import elbo_loss_tvae
from teras.utils.types import LayerOrModelType
from tqdm import tqdm


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class TVAE(keras.Model):
    """
    TVAE model with LayerFlow design.
    TVAE is a tabular data generation architecture proposed by the
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        input_dim: ``int``, dimensionality of the input dataset.
            The dimensionality of the input dataset.

            Note the dimensionality must be equal to the dimensionality of dataset
            that is passed to the fit method and not necessarily the dimensionality
            of the raw input dataset as sometimes data transformation alters the
            dimensionality of the dataset.

            You can access the dimensionality of the transformed dataset through the
            ``.data_dim`` attribute of the ``TVAEDataSampler`` instance used in sampling
            the dataset.

        metadata: ``dict``,
            A dictionary containing metadata for all features in the input data.
            This metadata is computed during the data transformation step and can be accessed
            from ``.get_metadata()`` method of ``TVAEDataTransformer`` instance.

        encoder: ``keras.layers.Layer`` or ``keras.models.Model``,
            An instance of ``TVAEEncoder`` or any custom keras layer/model that can work
            in its place.
            You can import the ``TVAEEncoder`` as follows,
                >>> from teras.models import TVAEEncoder

        decoder: ``keras.layers.Layer`` or ``keras.models.Model``,
            An instance of ``TVAEDecoder`` or any custom keras model/layer that can work
            in its palce.
            It decodes or decompresses from latent dimensions to data dimensions.
            You can import the ``TVAEDecoder`` as follows,
                >>> from teras.models import TVAEDecoder

        latent_dim: ``int``, default 128,
            Dimensionality of the learned latent space

        loss_factor: ``float``, default 2.,
            Hyperparameter used in the computation of ``ELBO loss`` for ``TVAE``.
            It controls how much the cross entropy loss contributes to the overall loss.
            It is directly proportional to the cross entropy loss.
    """
    def __init__(self,
                 encoder: LayerOrModelType,
                 decoder: LayerOrModelType,
                 latent_dim: int = 128,
                 loss_factor: float = 2.,
                 **kwargs):
        super().__init__(**kwargs)
        self.encoder = encoder
        self.decoder = decoder
        self.latent_dim = latent_dim
        self.loss_factor = loss_factor

    def call(self, inputs, training=None):
        mean, log_var, std = self.encoder(inputs)
        eps = tf.random.uniform(minval=0, maxval=1,
                                shape=tf.shape(std),
                                dtype=std.dtype)
        z = (std * eps) + mean
        generated_samples, sigmas = self.decoder(z)
        if training:
            loss = elbo_loss_tvae(data_dim=self.data_dim,
                                  real_samples=inputs,
                                  generated_samples=generated_samples,
                                  meta_data=self.meta_data,
                                  sigmas=sigmas,
                                  mean=mean,
                                  log_var=log_var,
                                  loss_factor=self.loss_factor)
            self.add_loss(loss)
            updated_simgas = tf.clip_by_value(self.decoder.sigmas,
                                              clip_value_min=0.01,
                                              clip_value_max=1.0)
            K.update(self.decoder.sigmas, updated_simgas)
        return generated_samples

    def generate(self,
                 num_samples: int,
                 data_transformer=None,
                 reverse_transform: bool = False,
                 batch_size: int = 512):
        """
        Generates new data using the trained Generator.

        Args:
            num_samples: ``int``,
                Number of new samples to generate

            data_transformer:
                Instance of ``TVAEDataTransformer`` class used to preprocess the raw data.
                This is required only if the `reverse_transform` is set to True.

            reverse_transform: ``bool``, default False,
                Whether to reverse transform the generated data to the original data format.
                If False, the raw generated data will be returned, which you can then manually
                transform into original data format by utilizing ``TVAEDataTransformer`` instance's
                ``reverse_transform`` method.

            batch_size: ``int``, default 512.
                If a value is passed, samples will be generated in batches
                where ``batch_size`` determines the size of each batch.
                If ``None``, all `num_samples` will be generated at once.
                Note that, if the number of samples to generate aren't huge
                or you know your hardware can handle to generate all samples at once,
                you can leave the value to None, otherwise it is recommended to specify
                a value for batch size.
        """
        if batch_size is None:
            batch_size = num_samples
        num_steps = num_samples // batch_size
        num_steps += 1 if num_samples % batch_size != 0 else 0
        generated_samples = []
        for _ in tqdm(range(num_steps), desc="Generating Data"):
            z = tf.random.normal(shape=[batch_size, self.latent_dim])
            gen_samples_temp, _ = self.decoder(z)
            generated_samples.append(gen_samples_temp)
        generated_samples = tf.concat(generated_samples, axis=0)
        generated_samples = generated_samples[:num_samples]

        if reverse_transform:
            if data_transformer is None:
                raise ValueError("""To reverse transform the raw generated data, `data_transformer` must not be None.
                              Please pass the instance of DataTransformer class used to transform the input
                              data. Or alternatively, you can set `reverse_transform` to False, and later
                              manually reverse transform the generated raw data to original format.""")
            generated_samples = data_transformer.reverse_transform(x_generated=generated_samples)
        return generated_samples

    def get_config(self):
        config = super().get_config()
        config.update({'encoder': keras.layers.serialize(self.encoder),
                       'decoder': keras.layers.serialize(self.decoder),
                       'latent_dim': self.latent_dim,
                       'loss_factor': self.loss_factor,
                       })
        return config

    @classmethod
    def from_config(cls, config):
        encoder = keras.layers.deserialize(config.pop("encoder"))
        decoder = keras.layers.deserialize(config.pop("decoder"))
        return cls(encoder=encoder,
                   decoder=decoder,
                   **config)



import tensorflow as tf
from tensorflow import keras


def generator_loss(generated_samples,
                   y_generated,
                   cond_vectors=None,
                   mask=None,
                   metadata=None):
    """
    Loss for the Generator model in the CTGAN architecture.

    CTGAN is a state-of-the-art tabular data generation architecture
    proposed by Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        generated_samples: Samples drawn from the input dataset
        y_generated: Discriminator's output for the generated samples
        cond_vectors: Conditional vectors that are used for and with
            generated samples
        mask: Mask created during the conditional vectors generation step
        metadata: Namedtuple meta deta of features.
            That meta data contains miscellaneous information about features,
            which is calculated during data transformation step.

    Returns:
        Generator's loss.
    """
    loss = []
    cross_entropy_loss = keras.losses.SparseCategoricalCrossentropy(from_logits=True,
                                                                    reduction=keras.losses.Reduction.NONE)
    numerical_features_relative_indices = metadata["numerical"]["relative_indices_all"]
    features_relative_indices_all = metadata["relative_indices_all"]
    num_categories_all = metadata["categorical"]["num_categories_all"]
    # the first k features in the data are numerical which we'll ignore as we're only
    # concerned with the categorical features here
    offset = len(numerical_features_relative_indices)
    for i, index in enumerate(features_relative_indices_all[offset:]):
        logits = generated_samples[:, index: index + num_categories_all[i]]
        temp_cond_vector = cond_vectors[:, i: i + num_categories_all[i]]
        labels = tf.argmax(temp_cond_vector, axis=1)
        ce_loss = cross_entropy_loss(y_pred=logits,
                                     y_true=labels
                                     )
        loss.append(ce_loss)
    loss = tf.stack(loss, axis=1)
    loss = tf.reduce_sum(loss * tf.cast(mask, dtype=tf.float32)) / tf.cast(tf.shape(y_generated)[0], dtype=tf.float32)
    loss = -tf.reduce_mean(y_generated) * loss
    return loss


def discriminator_loss(y_real, y_generated):
    """
    Loss for the Discriminator model in the CTGAN architecture.

    CTGAN is a state-of-the-art tabular data generation architecture
    proposed by Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        https://arxiv.org/abs/1907.00503

    Args:
        y_real: Discriminator's output for real samples
        y_generated: Discriminator's output for generated samples

    Returns:
        Discriminator's loss.
    """
    return -(tf.reduce_mean(y_real) - tf.reduce_mean(y_generated))


def generator_dummy_loss(y_dummy, y_pred):
    """
    For the generator model to track the loss function, and show it in outputs
    we create a dummy loss function which receives the loss function
    and returns it as is. It is passed to the model during compilation step.

    Reference(s):
        Idea taken from:
        https://towardsdatascience.com/solving-the-tensorflow-keras-model-loss-problem-fd8281aeeb11

    Args:
        y_dummy: An array of length batch_size, filled with dummy values.
        y_pred: The loss value computed using a custom loss function.

    Returns:
        Returns y_pred (i.e. loss) as is.
    """
    return tf.squeeze(y_pred)


from tensorflow import keras
from teras.layers.categorical_feature_embedding import CategoricalFeatureEmbedding
from teras.layers.saint.saint_numerical_feature_embedding import SAINTNumericalFeatureEmbedding
from teras.layers.saint.saint_encoder import SAINTEncoder
from teras.layers.common.head import ClassificationHead, RegressionHead
from teras.layerflow.models.saint import (SAINT as _SAINTLF,
                                          SAINTPretrainer as _SAINTPretrainerLF)
from teras.layers.saint.saint_reconstruction_head import SAINTReconstructionHead
from teras.layers.saint.saint_projection_head import SAINTProjectionHead
from teras.layers.regularization import MixUp, CutMix
from teras.config.saint import SAINTConfig
from teras.utils.types import UnitsValuesType


@keras.saving.register_keras_serializable("teras.models")
class SAINT(_SAINTLF):
    """
    SAINT architecture proposed by Gowthami Somepalli et al.
    in the paper,
    SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.

    SAINT performs attention over both rows and columns.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        features_metadata: ``dict``,
            a nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Embedding dimensions used in embedding numerical and categorical features.

        numerical_embedding_hidden_dim: ``int`` default 16,
            Dimensionality of the hidden layer that precedes the output layer in the
            ``SAINTNumericalFeatureEmebedding`` layer.

        num_transformer_layer: ``int``, default 6,
            Number of (SAINT) transformer layers to use in the Encoder.
            The encoder is used to contextualize the learned feature embeddings.

        num_attention_heads: ``int``, default 8,
            Number of attention heads to use in the ``MultiHeadSelfAttention`` layer
            that is part of the `Transformer` layer which in turn is part of the `Encoder`.

        num_inter_sample_attention_heads: ``int``, default 8,
            Number of heads to use in the ``MultiHeadInterSampleAttention`` that applies
            attention over rows.

        attention_dropout: ``float``, default 0.1, Dropout rate to use in the
            ``MultiHeadSelfAttention`` layer in the transformer layer.

        inter_sample_attention_dropout: `float`, default 0.1,
            Dropout rate for ``MultiHeadInterSampleAttention`` layer that applies
            attention over rows.

        feedforward_dropout: ``float``, default 0.1,
            Dropout rate to use for the dropout layer in the ``FeedForward`` layer.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multiplied with the `embedding_dim`
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the feedforward block.

        norm_epsilon: ``float``, default 1e-6,
            A very small number used for normalization in the ``LayerNormalization`` layer.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values.
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's string lookup layer.

        apply_attention_to_features: ``bool``, default True,
            Whether to apply attention over features using the regular ``MultiHeadAttenion`` layer.

        apply_attention_to_rows: ``bool``, default True,
            Whether to apply attention over rows using the SAINT ``MultiHeadInterSampleAttention``
            layer.
            Although it is strongly recommended to apply attention to both rows and features,
            but for experimentation's sake you can disable one of them, but NOT both at the
            same time!
    """
    def __init__(self,
                 input_dim: int,
                 features_metadata: dict,
                 embedding_dim: int = SAINTConfig.embedding_dim,
                 numerical_embedding_hidden_dim: int = SAINTConfig.numerical_embedding_hidden_dim,
                 num_transformer_layers: int = SAINTConfig.num_transformer_layers,
                 num_attention_heads: int = SAINTConfig.num_attention_heads,
                 num_inter_sample_attention_heads: int = SAINTConfig.num_inter_sample_attention_heads,
                 attention_dropout: float = SAINTConfig.attention_dropout,
                 inter_sample_attention_dropout: float = SAINTConfig.inter_sample_attention_dropout,
                 feedforward_dropout: float = SAINTConfig.feedforward_dropout,
                 feedforward_multiplier: int = SAINTConfig.feedforward_multiplier,
                 norm_epsilon: float = SAINTConfig.norm_epsilon,
                 encode_categorical_values: bool = SAINTConfig.encode_categorical_values,
                 apply_attention_to_features: bool = SAINTConfig.apply_attention_to_features,
                 apply_attention_to_rows: bool = SAINTConfig.apply_attention_to_rows,
                 **kwargs
                 ):
        num_categorical_features = len(features_metadata["categorical"])
        num_numerical_features = len(features_metadata["numerical"])
        numerical_features_exist = num_numerical_features > 0
        categorical_features_exist = num_categorical_features > 0

        # Numerical/Continuous Features Embedding
        numerical_feature_embedding = None
        if numerical_features_exist:
            # If numerical features exist, then they must be embedded
            numerical_feature_embedding = SAINTNumericalFeatureEmbedding(
                features_metadata=features_metadata,
                embedding_dim=embedding_dim,
                hidden_dim=numerical_embedding_hidden_dim
            )

        # Categorical Features Embedding
        categorical_feature_embedding = None
        if categorical_features_exist:
            # If categorical features exist, then they must be embedded
            categorical_feature_embedding = CategoricalFeatureEmbedding(
                features_metadata=features_metadata,
                embedding_dim=embedding_dim,
                encode=encode_categorical_values)

        encoder = SAINTEncoder(data_dim=input_dim,
                               num_transformer_layers=num_transformer_layers,
                               embedding_dim=embedding_dim,
                               num_attention_heads=num_attention_heads,
                               num_inter_sample_attention_heads=num_inter_sample_attention_heads,
                               attention_dropout=attention_dropout,
                               inter_sample_attention_dropout=inter_sample_attention_dropout,
                               feedforward_dropout=feedforward_dropout,
                               feedforward_multiplier=feedforward_multiplier,
                               norm_epsilon=norm_epsilon,
                               apply_attention_to_features=apply_attention_to_features,
                               apply_attention_to_rows=apply_attention_to_rows,
                               )

        super().__init__(input_dim=input_dim,
                         encoder=encoder,
                         categorical_feature_embedding=categorical_feature_embedding,
                         numerical_feature_embedding=numerical_feature_embedding,
                         **kwargs)

        self.features_metadata = features_metadata
        self.embedding_dim = embedding_dim
        self.num_transformer_layers = num_transformer_layers
        self.num_attention_heads = num_attention_heads
        self.num_inter_sample_attention_heads = num_inter_sample_attention_heads
        self.attention_dropout = attention_dropout
        self.inter_sample_attention_dropout = inter_sample_attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.feedforward_multiplier = feedforward_multiplier
        self.norm_epsilon = norm_epsilon
        self.encode_categorical_values = encode_categorical_values
        self.numerical_embedding_hidden_dim = numerical_embedding_hidden_dim
        self.apply_attention_to_features = apply_attention_to_features
        self.apply_attention_to_rows = apply_attention_to_rows

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'input_dim': self.input_dim,
                  'features_metadata': self.features_metadata,
                  'embedding_dim': self.embedding_dim,
                  'numerical_embedding_hidden_dim': self.numerical_embedding_hidden_dim,
                  'num_transformer_layers': self.num_transformer_layers,
                  'num_attention_heads': self.num_attention_heads,
                  'num_inter_sample_attention_heads': self.num_inter_sample_attention_heads,
                  'attention_dropout': self.attention_dropout,
                  'inter_sample_attention_dropout': self.inter_sample_attention_dropout,
                  'feedforward_dropout': self.feedforward_dropout,
                  'feedforward_multiplier': self.feedforward_multiplier,
                  'norm_epsilon': self.norm_epsilon,
                  'encode_categorical_values': self.encode_categorical_values,
                  'apply_attention_to_features': self.apply_attention_to_features,
                  'apply_attention_to_rows': self.apply_attention_to_rows,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        features_metadata = config.pop("features_metadata")
        return cls(input_dim=input_dim,
                   features_metadata=features_metadata, **config)


@keras.saving.register_keras_serializable("teras.models")
class SAINTClassifier(SAINT):
    """
    SAINTClassifier model based on the SAINT architecture proposed by
    Gowthami Somepalli et al. in the paper,
    SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.

    SAINT performs attention over both rows and columns.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        num_classes: ``int``, default 2,
            Number of classes to predict.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default [64, 32],
            Hidden units to use in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        features_metadata: ``dict``,
            a nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Embedding dimensions used in embedding numerical and categorical features.

        numerical_embedding_hidden_dim: ``int`` default 16,
            Dimensionality of the hidden layer that precedes the output layer in the
            ``SAINTNumericalFeatureEmebedding`` layer.

        num_transformer_layer: ``int``, default 6,
            Number of (SAINT) transformer layers to use in the Encoder.
            The encoder is used to contextualize the learned feature embeddings.

        num_attention_heads: ``int``, default 8,
            Number of attention heads to use in the ``MultiHeadSelfAttention`` layer
            that is part of the `Transformer` layer which in turn is part of the `Encoder`.

        num_inter_sample_attention_heads: ``int``, default 8,
            Number of heads to use in the ``MultiHeadInterSampleAttention`` that applies
            attention over rows.

        attention_dropout: ``float``, default 0.1, Dropout rate to use in the
            ``MultiHeadSelfAttention`` layer in the transformer layer.

        inter_sample_attention_dropout: `float`, default 0.1,
            Dropout rate for ``MultiHeadInterSampleAttention`` layer that applies
            attention over rows.

        feedforward_dropout: ``float``, default 0.1,
            Dropout rate to use for the dropout layer in the ``FeedForward`` layer.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multiplied with the `embedding_dim`
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the feedforward block.

        norm_epsilon: ``float``, default 1e-6,
            A very small number used for normalization in the ``LayerNormalization`` layer.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values.
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's string lookup layer.

        apply_attention_to_features: ``bool``, default True,
            Whether to apply attention over features using the regular ``MultiHeadAttenion`` layer.

        apply_attention_to_rows: ``bool``, default True,
            Whether to apply attention over rows using the SAINT ``MultiHeadInterSampleAttention``
            layer.
            Although it is strongly recommended to apply attention to both rows and features,
            but for experimentation's sake you can disable one of them, but NOT both at the
            same time!
    """
    def __init__(self,
                 num_classes: int = 2,
                 head_units_values: UnitsValuesType = (64, 32),
                 input_dim: int = None,
                 features_metadata: dict = None,
                 embedding_dim: int = SAINTConfig.embedding_dim,
                 numerical_embedding_hidden_dim: int = SAINTConfig.numerical_embedding_hidden_dim,
                 num_transformer_layers: int = SAINTConfig.num_transformer_layers,
                 num_attention_heads: int = SAINTConfig.num_attention_heads,
                 num_inter_sample_attention_heads: int = SAINTConfig.num_inter_sample_attention_heads,
                 attention_dropout: float = SAINTConfig.attention_dropout,
                 inter_sample_attention_dropout: float = SAINTConfig.inter_sample_attention_dropout,
                 feedforward_dropout: float = SAINTConfig.feedforward_dropout,
                 feedforward_multiplier: int = SAINTConfig.feedforward_multiplier,
                 norm_epsilon: float = SAINTConfig.norm_epsilon,
                 encode_categorical_values: bool = SAINTConfig.encode_categorical_values,
                 apply_attention_to_features: bool = SAINTConfig.apply_attention_to_features,
                 apply_attention_to_rows: bool = SAINTConfig.apply_attention_to_rows,
                 **kwargs
                 ):
        if features_metadata is None:
            raise ValueError("`features_metadata` cannot be None. "
                             "You can get features metadata dictionary using the `get_features_metadata_for_embedding` "
                             "function from teras.utils.")
        # Since the parent `SAINT` class subclasses its LayerFlow version
        # which accepts layers as input and does expose a head argument to pass the layer.
        # we can create a relevant head layer, in this case, the `ClassificationHead` layer
        # and pass its instance to the parent who will pass it along to its parent as part of
        # the **kwargs dict
        head = ClassificationHead(num_classes=num_classes,
                                  units_values=head_units_values,
                                  activation_hidden="relu",
                                  normalization="batch",
                                  name="saint_classifier_head")
        super().__init__(input_dim=input_dim,
                         features_metadata=features_metadata,
                         embedding_dim=embedding_dim,
                         numerical_embedding_hidden_dim=numerical_embedding_hidden_dim,
                         num_transformer_layers=num_transformer_layers,
                         num_attention_heads=num_attention_heads,
                         num_inter_sample_attention_heads=num_inter_sample_attention_heads,
                         attention_dropout=attention_dropout,
                         inter_sample_attention_dropout=inter_sample_attention_dropout,
                         feedforward_dropout=feedforward_dropout,
                         feedforward_multiplier=feedforward_multiplier,
                         norm_epsilon=norm_epsilon,
                         encode_categorical_values=encode_categorical_values,
                         apply_attention_to_features=apply_attention_to_features,
                         apply_attention_to_rows=apply_attention_to_rows,
                         head=head,
                         **kwargs)
        self.num_classes = num_classes
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        config.update({'num_classes': self.num_classes,
                       'head_units_values': self.head_units_values
                       })
        return config


@keras.saving.register_keras_serializable("teras.models")
class SAINTRegressor(SAINT):
    """
    SAINTRegressor model based on the SAINT architecture proposed by
    Gowthami Somepalli et al. in the paper,
    SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.

    SAINT performs attention over both rows and columns.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        num_outputs: ``int``, default 1,
            Number of regression outputs to predict.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default [64, 32],
            Hidden units to use in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        features_metadata: ``dict``,
            a nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Embedding dimensions used in embedding numerical and categorical features.

        numerical_embedding_hidden_dim: ``int`` default 16,
            Dimensionality of the hidden layer that precedes the output layer in the
            ``SAINTNumericalFeatureEmebedding`` layer.

        num_transformer_layer: ``int``, default 6,
            Number of (SAINT) transformer layers to use in the Encoder.
            The encoder is used to contextualize the learned feature embeddings.

        num_attention_heads: ``int``, default 8,
            Number of attention heads to use in the ``MultiHeadSelfAttention`` layer
            that is part of the `Transformer` layer which in turn is part of the `Encoder`.

        num_inter_sample_attention_heads: ``int``, default 8,
            Number of heads to use in the ``MultiHeadInterSampleAttention`` that applies
            attention over rows.

        attention_dropout: ``float``, default 0.1, Dropout rate to use in the
            ``MultiHeadSelfAttention`` layer in the transformer layer.

        inter_sample_attention_dropout: `float`, default 0.1,
            Dropout rate for ``MultiHeadInterSampleAttention`` layer that applies
            attention over rows.

        feedforward_dropout: ``float``, default 0.1,
            Dropout rate to use for the dropout layer in the ``FeedForward`` layer.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multiplied with the `embedding_dim`
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the feedforward block.

        norm_epsilon: ``float``, default 1e-6,
            A very small number used for normalization in the ``LayerNormalization`` layer.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values.
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's string lookup layer.

        apply_attention_to_features: ``bool``, default True,
            Whether to apply attention over features using the regular ``MultiHeadAttenion`` layer.

        apply_attention_to_rows: ``bool``, default True,
            Whether to apply attention over rows using the SAINT ``MultiHeadInterSampleAttention``
            layer.
            Although it is strongly recommended to apply attention to both rows and features,
            but for experimentation's sake you can disable one of them, but NOT both at the
            same time!
    """

    def __init__(self,
                 num_outputs: int = 1,
                 head_units_values: UnitsValuesType = (64, 32),
                 input_dim: int = None,
                 features_metadata: dict = None,
                 embedding_dim: int = SAINTConfig.embedding_dim,
                 numerical_embedding_hidden_dim: int = SAINTConfig.numerical_embedding_hidden_dim,
                 num_transformer_layers: int = SAINTConfig.num_transformer_layers,
                 num_attention_heads: int = SAINTConfig.num_attention_heads,
                 num_inter_sample_attention_heads: int = SAINTConfig.num_inter_sample_attention_heads,
                 attention_dropout: float = SAINTConfig.attention_dropout,
                 inter_sample_attention_dropout: float = SAINTConfig.inter_sample_attention_dropout,
                 feedforward_dropout: float = SAINTConfig.feedforward_dropout,
                 feedforward_multiplier: int = SAINTConfig.feedforward_multiplier,
                 norm_epsilon: float = SAINTConfig.norm_epsilon,
                 encode_categorical_values: bool = SAINTConfig.encode_categorical_values,
                 apply_attention_to_features: bool = SAINTConfig.apply_attention_to_features,
                 apply_attention_to_rows: bool = SAINTConfig.apply_attention_to_rows,
                 **kwargs
                 ):
        if features_metadata is None:
            raise ValueError("`features_metadata` cannot be None. "
                             "You can get features metadata dictionary using the `get_features_metadata_for_embedding` "
                             "function from teras.utils.")
        head = RegressionHead(num_outputs=num_outputs,
                              units_values=head_units_values,
                              activation_hidden="relu",
                              normalization="batch",
                              name="saint_regressor_head")
        super().__init__(input_dim=input_dim,
                         features_metadata=features_metadata,
                         embedding_dim=embedding_dim,
                         numerical_embedding_hidden_dim=numerical_embedding_hidden_dim,
                         num_transformer_layers=num_transformer_layers,
                         num_attention_heads=num_attention_heads,
                         num_inter_sample_attention_heads=num_inter_sample_attention_heads,
                         attention_dropout=attention_dropout,
                         inter_sample_attention_dropout=inter_sample_attention_dropout,
                         feedforward_dropout=feedforward_dropout,
                         feedforward_multiplier=feedforward_multiplier,
                         norm_epsilon=norm_epsilon,
                         encode_categorical_values=encode_categorical_values,
                         apply_attention_to_features=apply_attention_to_features,
                         apply_attention_to_rows=apply_attention_to_rows,
                         head=head,
                         **kwargs)
        self.num_outputs = num_outputs
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        config.update({'num_outputs': self.num_outputs,
                       'head_units_values': self.head_units_values,
                       })
        return config


@keras.saving.register_keras_serializable("teras.models")
class SAINTPretrainer(_SAINTPretrainerLF):
    """
    SAINTPretrainer model based on the pretraining architecture
    for the SAINT model proposed by Gowthami Somepalli et al.
    in the paper,
    SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.

    SAINT performs attention over both rows and columns.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        model: ``SAINT``,
            An instance of the ``SAINT`` model that you want to pretrain.
            Note that, you should use the base ``SAINT`` model's instance,
            not ``SAINTClassifier`` or ``SAINTRegressor``.
            Using default API, you can import it as,
                >>> from teras.models import SAINT
            Using LayerFlow API, you can import it as,
                >>> from teras.layerflow.models import SAINT
                And REMEMBER to leave the ``head`` argment as None.

        features_metadata: ``dict``,
            a nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Embedding dimensions being used by the base model for embedding
            numerical and categorical features.

        cutmix_probs: ``float``, default 0.1,
            ``CutMix`` probability which is used in generation of mask
            that is used to mix samples together.

        mixup_alpha: ``float``, default 1.0,
            Alpha value for the ``MixUp`` layer, that is used for the
            Beta distribution to sample `lambda_`
            which is used to interpolate samples.

        temperature: ``float``, default 0.7,
            Temperature value used in the computation of the InfoNCE contrastive loss.

        lambda_: ``float``, default 10,
            Controls the weightage of denoising loss in the summation of denoising and
            contrastive loss.
    """
    def __init__(self,
                 model: SAINT,
                 features_metadata: dict,
                 embedding_dim: int = 32,
                 cutmix_probs: float = 0.3,
                 mixup_alpha: float = 1.0,
                 temperature: float = 0.7,
                 lambda_: float = 10.,
                 **kwargs):
        mixup = MixUp(alpha=mixup_alpha)
        cutmix = CutMix(probs=cutmix_probs)
        num_features = len(features_metadata["categorical"]) + len(features_metadata["numerical"])

        # For the computation of contrastive loss, we use projection heads.
        # Projection head hidden dimensions as calculated by the
        # official implementation
        projection_head_hidden_dim = 6 * embedding_dim * num_features // 5
        projection_head_output_dim = embedding_dim * num_features // 2

        projection_head_1 = SAINTProjectionHead(hidden_dim=projection_head_hidden_dim,
                                                output_dim=projection_head_output_dim,
                                                name="projection_head_for_original_data")

        projection_head_2 = SAINTProjectionHead(hidden_dim=projection_head_hidden_dim,
                                                output_dim=projection_head_output_dim,
                                                name="projection_head_for_augmented_data")

        reconstruction_head = SAINTReconstructionHead(features_metadata=model.features_metadata,
                                                      embedding_dim=model.embedding_dim)

        super().__init__(model=model,
                         features_metadata=features_metadata,
                         mixup=mixup,
                         cutmix=cutmix,
                         projection_head_1=projection_head_1,
                         projection_head_2=projection_head_2,
                         reconstruction_head=reconstruction_head,
                         **kwargs)
        self.model = model
        self.features_metadata = features_metadata
        self.embedding_dim = embedding_dim
        self.cutmix_probs = cutmix_probs
        self.mixup_alpha = mixup_alpha
        self.temperature = temperature
        self.lambda_ = lambda_

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'model': keras.layers.serialize(self.model),
                  'features_metadata': self.features_metadata,
                  'embedding_dim': self.embedding_dim,
                  'cutmix_probs': self.cutmix_probs,
                  'mixup_alpha': self.mixup_alpha,
                  'temperature': self.temperature,
                  'lambda_': self.lambda_,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        model = keras.layers.deserialize(config.pop("model"))
        features_metadata = config.pop("features_metadata")
        return cls(model, features_metadata, **config)



import tensorflow as tf
from teras.layers.preprocessing.periodic_embedding import PeriodicEmbedding
from teras.utils import get_features_metadata_for_embedding
import pandas as pd
import pytest


@pytest.fixture
def setup_data():
    df = pd.DataFrame({'player_level': [70, 90, 80, 90, 100, 90, 70, 80],
                       'shirt_number': [10, 7, 7, 10, 10, 10, 7, 10],
                       'income': [1000, 1270, 900, 1000, 1234, 5678, 9999, 0],
                       'goals': [900, 70, 222, 131, 1337, 99, 0, 656]})
    metadata = get_features_metadata_for_embedding(df,
                                                   categorical_features=['player_level', 'shirt_number'],
                                                   numerical_features=['income', 'goals'])
    return df.values, metadata


def test_periodic_embedding_valid_call(setup_data):
    inputs, metadata = setup_data
    periodic_embedding = PeriodicEmbedding(features_metadata=metadata,
                                           embedding_dim=16)
    outputs = periodic_embedding(inputs)


def test_periodic_embedding_output_shape(setup_data):
    inputs, metadata = setup_data
    periodic_embedding = PeriodicEmbedding(features_metadata=metadata,
                                           embedding_dim=16)
    outputs = periodic_embedding(inputs)
    assert len(tf.shape(outputs)) == 3
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 2    # PeriodicEmbedding halves the features
    assert tf.shape(outputs)[2] == 16


def test_periodic_embedding_raises_error_when_numerical_features_dont_exist():
    df = pd.DataFrame({'player_level': [70, 90, 80, 90, 100, 90, 70, 80],
                       'shirt_number': [10, 7, 7, 10, 10, 10, 7, 10]})
    metadata = get_features_metadata_for_embedding(df,
                                                   categorical_features=['player_level', 'shirt_number'],
                                                   numerical_features=None)
    with pytest.raises(ValueError):
        extractor = PeriodicEmbedding(features_metadata=metadata)



import tensorflow as tf
from tensorflow import keras


# ================================= GLU =======================================
@keras.saving.register_keras_serializable(package="teras.activations")
def glu(logits: tf.Tensor) -> tf.Tensor:
    """
    Generalized linear unit nonlinear activation.

    logits: ``Tensor``,
        Input tensor of logits.
    """
    x, gates = tf.split(logits,
                        num_or_size_splits=2,
                        axis=-1)
    return x * tf.nn.sigmoid(gates)


# ================================= GEGLU =======================================
@keras.saving.register_keras_serializable(package="teras.activations")
def geglu(logits: tf.Tensor) -> tf.Tensor:
    """
    GeGLU is an activation function which is a variant of GLU

    logits: ``Tensor``,
        Input tensor of logits.
    """
    x, gates = tf.split(logits,
                        num_or_size_splits=2,
                        axis=-1)
    return x * tf.nn.gelu(gates)


# ================================= SPARSEMAX =======================================
# The function below is copied from TensorFlow Addons with slight modifications.
# And reason we copied it here because TensorFlow Addons has ended development.
# RIP o7
@keras.saving.register_keras_serializable(package="teras.activations")
def sparsemax(logits: tf, axis: int = -1) -> tf.Tensor:
    """
    Sparsemax activation function.

    Referencne(s):
        https://arxiv.org/abs/1602.02068.

    Usage:

    >>> x = tf.constant([[-1.0, 0.0, 1.0], [-5.0, 1.0, 2.0]])
    >>> teras.activations.sparsemax(x)
    <tf.Tensor: shape=(2, 3), dtype=float32, numpy=
    array([[0., 0., 1.],
           [0., 0., 1.]], dtype=float32)>

    Args:
        logits: ``Tensor``.
        axis: ``int``, axis along which the sparsemax operation is applied.
    Returns:
        A `Tensor`, output of sparsemax transformation. Has the same type and
        shape as `logits`.
    Raises:
        ValueError: In case `dim(logits) == 1`.
    """
    logits = tf.convert_to_tensor(logits, name="logits")

    # We need its original shape for shape inference.
    shape = logits.get_shape()
    rank = shape.rank
    is_last_axis = (axis == -1) or (axis == rank - 1)

    if is_last_axis:
        output = _compute_2d_sparsemax(logits)
        output.set_shape(shape)
        return output

    # If dim is not the last dimension, we have to do a transpose so that we can
    # still perform softmax on its last dimension.

    # Swap logits' dimension of dim and its last dimension.
    rank_op = tf.rank(logits)
    axis_norm = axis % rank
    logits = _swap_axis(logits, axis_norm, tf.math.subtract(rank_op, 1))

    # Do the actual softmax on its last dimension.
    output = _compute_2d_sparsemax(logits)
    output = _swap_axis(output, axis_norm, tf.math.subtract(rank_op, 1))

    # Make shape inference work since transpose may erase its static shape.
    output.set_shape(shape)
    return output


def _swap_axis(logits, dim_index, last_index, **kwargs):
    return tf.transpose(
        logits,
        tf.concat(
            [
                tf.range(dim_index),
                [last_index],
                tf.range(dim_index + 1, last_index),
                [dim_index],
            ],
            0,
        ),
        **kwargs,
    )


def _compute_2d_sparsemax(logits):
    """Performs the sparsemax operation when axis=-1."""
    shape_op = tf.shape(logits)
    obs = tf.math.reduce_prod(shape_op[:-1])
    dims = shape_op[-1]

    # In the paper, they call the logits z.
    # The mean(logits) can be substracted from logits to make the algorithm
    # more numerically stable. the instability in this algorithm comes mostly
    # from the z_cumsum. Substacting the mean will cause z_cumsum to be close
    # to zero. However, in practise the numerical instability issues are very
    # minor and substacting the mean causes extra issues with inf and nan
    # input.
    # Reshape to [obs, dims] as it is almost free and means the remanining
    # code doesn't need to worry about the rank.
    z = tf.reshape(logits, [obs, dims])

    # sort z
    z_sorted, _ = tf.nn.top_k(z, k=dims)

    # calculate k(z)
    z_cumsum = tf.math.cumsum(z_sorted, axis=-1)
    k = tf.range(1, tf.cast(dims, logits.dtype) + 1, dtype=logits.dtype)
    z_check = 1 + k * z_sorted > z_cumsum
    # because the z_check vector is always [1,1,...1,0,0,...0] finding the
    # (index + 1) of the last `1` is the same as just summing the number of 1.
    k_z = tf.math.reduce_sum(tf.cast(z_check, tf.int32), axis=-1)

    # calculate tau(z)
    # If there are inf values or all values are -inf, the k_z will be zero,
    # this is mathematically invalid and will also cause the gather_nd to fail.
    # Prevent this issue for now by setting k_z = 1 if k_z = 0, this is then
    # fixed later (see p_safe) by returning p = nan. This results in the same
    # behavior as softmax.
    k_z_safe = tf.math.maximum(k_z, 1)
    indices = tf.stack([tf.range(0, obs), tf.reshape(k_z_safe, [-1]) - 1], axis=1)
    tau_sum = tf.gather_nd(z_cumsum, indices)
    tau_z = (tau_sum - 1) / tf.cast(k_z, logits.dtype)

    # calculate p
    p = tf.math.maximum(tf.cast(0, logits.dtype), z - tf.expand_dims(tau_z, -1))
    # If k_z = 0 or if z = nan, then the input is invalid
    p_safe = tf.where(
        tf.expand_dims(
            tf.math.logical_or(tf.math.equal(k_z, 0), tf.math.is_nan(z_cumsum[:, -1])),
            axis=-1,
        ),
        tf.fill([obs, dims], tf.cast(float("nan"), logits.dtype)),
        p,
    )

    # Reshape back to original size
    p_safe = tf.reshape(p_safe, shape_op)
    return p_safe


# ================================= GUMBLE SOFTMAX =======================================
@keras.saving.register_keras_serializable(package="teras.activations")
def gumbel_softmax(logits: tf.Tensor,
                   temperature: float = 0.2,
                   hard: bool = False) -> tf.Tensor:
    """
    Implementation of the Gumbel Softmax activation function
    proposed by Eric Jang et al. in the paper
    Categorical Reparameterization with Gumbel-Softmax

    Reference(s):
        https://arxiv.org/abs/1611.01144

    Args:
        logits: ``Tensor`` or ``ndarray``
            Input tensor of logits.

        temperature: ``float``, default 0.2,
            Controls the sharpness or smoothness of the resulting probability distribution.
            A higher temperature value leads to a smoother and more uniform probability distribution.
            Conversely, a lower temperature value makes the distribution concentrated around
            the category with the highest probability.

        hard: ``bool``, default False,
            Whether to return soft probabilities or hard one hot vectors.
    """
    u = tf.random.uniform(tf.shape(logits),
                          minval=0,
                          maxval=1)
    gumbels = -tf.math.log(-tf.math.log(u))
    perturbed_logits = (logits + gumbels) / temperature
    probabilities = tf.nn.softmax(perturbed_logits)
    if hard:
        one_hot_labels = tf.one_hot(tf.argmax(probabilities, axis=-1),
                                    tf.shape(logits)[-1])
        return one_hot_labels
    return probabilities



import tensorflow as tf
from tensorflow import keras
from teras.layers.tabnet import TabNetFeatureTransformer


@keras.saving.register_keras_serializable("teras.layers.tabnet")
class TabNetDecoder(keras.layers.Layer):
    """
    TabNetDecoder as proposed by Sercan et al. in TabNet paper.

    Reference(s):
        https://arxiv.org/abs/1908.07442

    Args:
        data_dim: ``int``,
            The dimensionality of the original input dataset,
            or the number of features in the original input dataset.

        feature_transformer_dim: ``int``, default 32,
            The dimensionality of the hidden representation in feature transformation block.
            Each layer first maps the representation to a `2 * feature_transformer_dim`
            output and half of it is used to determine the
            non-linearity of the GLU activation where the other half is used as an
            input to GLU, and eventually `feature_transformer_dim` output is
            transferred to the next layer.

        decision_step_output_dim: ``int``, default 32,
            The dimensionality of output at each decision step, which is later mapped to the
            final classification or regression output.
            It is recommended to keep ``decision_step_output_dim`` and ``feature_transformer_dim``
            equal to each other.
            Adjusting these two parameters values is a good way of obtaining a tradeoff between
            performance and complexity.

        num_decision_steps: ``int``, default 5,
            The number of sequential decision steps.
            For most datasets a value in the range [3, 10] is optimal.
            If there are more informative features in the dataset, the value tends to
            be higher. That said, a very high value of `num_decision_steps` may suffer
            from overfitting.

        num_shared_layers: ``int``, default 2,
            Number of shared layers to use in the ``TabNetFeatureTransformer``.
            These shared layers are *shared* across decision steps.

        num_decision_dependent_layers: ``int``, default 2,
            Number of decision dependent layers to use in the ``TabNetFeatureTransformer``.
            In simple words, ``num_decision_dependent_layers`` are created
            for each decision step in the ``num_decision_steps``.
            For instance, if ``num_decision_steps = `5` and  ``num_decision_dependent_layers = 2``
            then 10 layers will be created, 2 for each decision step.

        relaxation_factor: ``float``, default 1.5,
            Relaxation factor that promotes the reuse of each
            feature at different decision steps. When it is 1, a feature is enforced
            to be used only at one decision step and as it increases, more
            flexibility is provided to use a feature at multiple decision steps.
            An optimal value of relaxation_factor can have a major role on the performance.
            Typically, a larger value for `num_decision_steps` favors for a larger ``relaxation_factor``.

        batch_momentum: ``float``, default 0.9,
            Momentum value to use for ``BatchNormalization`` layer.

        virtual_batch_size: `int`, default 64,
            Batch size to use for ``virtual_batch_size`` parameter in ``BatchNormalization`` layer.
            This is typically much smaller than the ``batch_size`` used for training.

        residual_normalization_factor: ``float``, default 0.5,
            In the feature transformer, except for the first layer, every other layer utilizes
            normalized residuals, where ``residual_normalization_factor``
            determines the scale of normalization.
    """
    def __init__(self,
                 data_dim: int,
                 feature_transformer_dim: int = 32,
                 decision_step_output_dim: int = 32,
                 num_decision_steps: int = 5,
                 num_shared_layers: int = 2,
                 num_decision_dependent_layers: int = 2,
                 batch_momentum: float = 0.9,
                 virtual_batch_size: int = 64,
                 residual_normalization_factor: float = 0.5,
                 **kwargs):
        super().__init__(**kwargs)
        self.data_dim = data_dim
        self.feature_transformer_dim = feature_transformer_dim
        self.decision_step_output_dim = decision_step_output_dim
        self.num_shared_layers = num_shared_layers
        self.num_decision_dependent_layers = num_decision_dependent_layers
        self.num_decision_steps = num_decision_steps
        self.batch_momentum = batch_momentum
        self.virtual_batch_size = virtual_batch_size
        self.residual_normalization_factor = residual_normalization_factor

        self.features_transformers_per_step = []
        self.projection_layers_per_step = []

        # To be able to share the `shared_layers` across instances, we introduced
        # a class attribute called `shared_layers` in the FeatureTransformer class
        # BUT here's the problem,
        # even though we now want to create new shared layers separately for our Decoder, it won't ... unless...
        # we do this
        TabNetFeatureTransformer.reset_shared_layers()
        # Why? Because each time we create `TabNetFeatureTransformer` instance, it first checks if `shared_layers`
        # attribute is None or not, if it's None it will create new shared layers, otherwise it won't.

        for i in range(self.num_decision_steps):
            self.features_transformers_per_step.append(TabNetFeatureTransformer(
                                                        units=self.feature_transformer_dim,
                                                        num_shared_layers=self.num_shared_layers,
                                                        num_decision_dependent_layers=self.num_decision_dependent_layers,
                                                        batch_momentum=self.batch_momentum,
                                                        virtual_batch_size=self.virtual_batch_size,
                                                        residual_normalization_factor=self.residual_normalization_factor,
                                                        name=f"step_{i}_feature_transformer"
                                                        ))
            self.projection_layers_per_step.append(keras.layers.Dense(self.data_dim))

    def call(self, inputs, mask=None):
        """
        Args:
            inputs: Encoded representations.

        Returns:

        """
        batch_size = tf.shape(inputs)[0]
        reconstructed_features = tf.zeros(shape=(batch_size, self.data_dim))

        for i in range(self.num_decision_steps):
            feat_output = self.features_transformers_per_step[i](inputs)
            reconstructed_features += self.projection_layers_per_step[i](feat_output)

        # The paper says,
        # the decoder’s last FC (dense) layer is multiplied with S (binary mask indicating which features are missing)
        # to output the unknown features.
        if mask is not None:
            reconstructed_features *= mask

        return reconstructed_features

    def get_config(self):
        config = super().get_config()
        new_config = {'data_dim': self.data_dim,
                      'feature_transformer_dim': self.feature_transformer_dim,
                      'decision_step_output_dim': self.decision_step_output_dim,
                      'num_decision_steps': self.num_decision_steps,
                      'num_shared_layers': self.num_shared_layers,
                      'num_decision_dependent_layers': self.num_decision_dependent_layers,
                      'batch_momentum': self.batch_momentum,
                      'virtual_batch_size': self.virtual_batch_size,
                      'residual_normalization_factor': self.residual_normalization_factor,
                      }
        config.update(new_config)
        return config



from tensorflow import keras
from teras.layers.ft_transformer.ft_numerical_feature_embedding import FTNumericalFeatureEmbedding
from teras.layers.ft_transformer.ft_cls_token import FTCLSToken
from teras.layers.categorical_feature_embedding import CategoricalFeatureEmbedding
from teras.layers.common.transformer import Encoder
from teras.layers.common.head import ClassificationHead, RegressionHead
from teras.layerflow.models.ft_transformer import FTTransformer as _FTTransformerLF
from teras.utils.types import UnitsValuesType


@keras.saving.register_keras_serializable("teras.models")
class FTTransformer(_FTTransformerLF):
    """
    FT Transformer architecture proposed by Yury Gorishniy et al.
    in the paper Revisiting Deep Learning Models for Tabular Data
    in their FTTransformer architecture.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Embedding dimensions used in embedding numerical and categorical features.

        numerical_embedding_hidden_dim: ``int`` default 16,
            Dimensionality of the hidden layer that precedes the output layer in the
            SAINT NumericalFeatureEmebedding layer.

        num_transformer_layer: ``int``, default 6,
            Number of ``Transformer`` layers to use in the ``Encoder``.
            The encoder is used to contextualize the learned feature embeddings.

        num_attention_heads: ``int``, default 8,
            Number of attention heads to use in the ``MultiHeadSelfAttention`` layer
            that is part of the `Transformer` layer which in turn is part of the `Encoder`.

        attention_dropout: ``float``, default 0.1,
            Dropout rate to use in the ``MultiHeadSelfAttention`` layer in the transformer layer.

        feedforward_dropout: ``float``, default 0.1,
            Dropout rate to use for the dropout layer in the FeedForward block.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multipled with the `embedding_dim`
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the feedforward block.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values.
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's ``IntegerLookup`` layer.
    """
    def __init__(self,
                 input_dim: int,
                 features_metadata: dict,
                 embedding_dim: int = 32,
                 num_transformer_layers: int = 8,
                 num_attention_heads: int = 8,
                 attention_dropout: float = 0.1,
                 feedforward_dropout:  float = 0.05,
                 feedforward_multiplier: int = 4,
                 encode_categorical_values: bool = True,
                 **kwargs):
        num_categorical_features = len(features_metadata["categorical"])
        num_numerical_features = len(features_metadata["numerical"])
        categorical_features_exist = num_categorical_features > 0
        numerical_features_exist = num_numerical_features > 0

        # Numerical/Continuous Features Embedding
        numerical_feature_embedding = None
        if numerical_features_exist:
            numerical_feature_embedding = FTNumericalFeatureEmbedding(features_metadata=features_metadata,
                                                                      embedding_dim=embedding_dim)

        # Categorical Features Embedding
        categorical_feature_embedding = None
        if categorical_features_exist:
            # If categorical features exist, then they must be embedded
            categorical_feature_embedding = CategoricalFeatureEmbedding(
                features_metadata=features_metadata,
                embedding_dim=embedding_dim,
                encode=encode_categorical_values)

        cls_token = FTCLSToken(embedding_dim,
                               initialization="normal")
        encoder = Encoder(num_transformer_layers=num_transformer_layers,
                          num_attention_heads=num_attention_heads,
                          embedding_dim=embedding_dim,
                          attention_dropout=attention_dropout,
                          feedforward_dropout=feedforward_dropout,
                          feedforward_multiplier=feedforward_multiplier)

        super().__init__(input_dim=input_dim,
                         categorical_feature_embedding=categorical_feature_embedding,
                         numerical_feature_embedding=numerical_feature_embedding,
                         cls_token=cls_token,
                         encoder=encoder,
                         **kwargs
                         )
        self.input_dim = input_dim
        self.features_metadata = features_metadata
        self.num_transformer_layers = num_transformer_layers
        self.embedding_dim = embedding_dim
        self.num_attention_heads = num_attention_heads
        self.attention_dropout = attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.feedforward_multiplier = feedforward_multiplier
        self.encode_categorical_values = encode_categorical_values

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'input_dim': self.input_dim,
                  'features_metadata': self.features_metadata,
                  'embedding_dim': self.embedding_dim,
                  'num_transformer_layers': self.num_transformer_layers,
                  'num_attention_heads': self.num_attention_heads,
                  'attention_dropout': self.attention_dropout,
                  'feedforward_dropout': self.feedforward_dropout,
                  'feedforward_multiplier': self.feedforward_multiplier,
                  'encode_categorical_values': self.encode_categorical_values}
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        features_metadata = config.pop("features_metadata")
        return cls(input_dim=input_dim,
                   features_metadata=features_metadata,
                   **config)


@keras.saving.register_keras_serializable("teras.models")
class FTTransformerClassifier(FTTransformer):
    """
    FTTransformerClassifier based on the FTTransformer architecture proposed
    by Yury Gorishniy et al. in the paper,
    Revisiting Deep Learning Models for Tabular Data.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        num_classes: ``int``, default 2,
            Number of classes to predict.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default None,
            Units values to use in the hidden layers in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.
            By default, no hidden layer is used.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Embedding dimensions used in embedding numerical and categorical features.

        numerical_embedding_hidden_dim: ``int`` default 16,
            Dimensionality of the hidden layer that precedes the output layer in the
            SAINT NumericalFeatureEmebedding layer.

        num_transformer_layer: ``int``, default 6,
            Number of ``Transformer`` layers to use in the ``Encoder``.
            The encoder is used to contextualize the learned feature embeddings.

        num_attention_heads: ``int``, default 8,
            Number of attention heads to use in the ``MultiHeadSelfAttention`` layer
            that is part of the `Transformer` layer which in turn is part of the `Encoder`.

        attention_dropout: ``float``, default 0.1,
            Dropout rate to use in the ``MultiHeadSelfAttention`` layer in the transformer layer.

        feedforward_dropout: ``float``, default 0.1,
            Dropout rate to use for the dropout layer in the FeedForward block.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multipled with the `embedding_dim`
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the feedforward block.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values.
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's ``IntegerLookup`` layer.
    """
    def __init__(self,
                 num_classes: int = 2,
                 head_units_value: UnitsValuesType = None,
                 features_metadata: dict = None,
                 input_dim: int = None,
                 embedding_dim: int = 32,
                 num_transformer_layers: int = 8,
                 num_attention_heads: int = 8,
                 attention_dropout: float = 0.1,
                 feedforward_dropout:  float = 0.05,
                 feedforward_multiplier: int = 4,
                 encode_categorical_values: bool = True,
                 **kwargs):
        head = ClassificationHead(num_classes=num_classes,
                                  units_values=head_units_value,
                                  normalization="layer")
        super().__init__(input_dim=input_dim,
                         features_metadata=features_metadata,
                         embedding_dim=embedding_dim,
                         num_transformer_layers=num_transformer_layers,
                         num_attention_heads=num_attention_heads,
                         attention_dropout=attention_dropout,
                         feedforward_dropout=feedforward_dropout,
                         feedforward_multiplier=feedforward_multiplier,
                         encode_categorical_values=encode_categorical_values,
                         head=head,
                         **kwargs)
        self.num_classes = num_classes
        self.head_units_value = head_units_value

    def get_config(self):
        config = super().get_config()
        config.update({'num_classes': self.num_classes,
                       'head_units_value': self.head_units_value})
        return config


@keras.saving.register_keras_serializable("teras.models")
class FTTransformerRegressor(FTTransformer):
    """
    FTTransformerRegressor based on the FTTransformer architecture proposed
    by Yury Gorishniy et al. in the paper,
    Revisiting Deep Learning Models for Tabular Data.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        num_outputs: `int`, default 1,
            Number of outputs to predict.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default None,
            Units values to use in the hidden layers in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.
            By default, no hidden layer is used.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Embedding dimensions used in embedding numerical and categorical features.

        numerical_embedding_hidden_dim: ``int`` default 16,
            Dimensionality of the hidden layer that precedes the output layer in the
            SAINT NumericalFeatureEmebedding layer.

        num_transformer_layer: ``int``, default 6,
            Number of ``Transformer`` layers to use in the ``Encoder``.
            The encoder is used to contextualize the learned feature embeddings.

        num_attention_heads: ``int``, default 8,
            Number of attention heads to use in the ``MultiHeadSelfAttention`` layer
            that is part of the `Transformer` layer which in turn is part of the `Encoder`.

        attention_dropout: ``float``, default 0.1,
            Dropout rate to use in the ``MultiHeadSelfAttention`` layer in the transformer layer.

        feedforward_dropout: ``float``, default 0.1,
            Dropout rate to use for the dropout layer in the FeedForward block.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multipled with the `embedding_dim`
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the feedforward block.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values.
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's ``IntegerLookup`` layer.
    """

    def __init__(self,
                 num_outputs: int = 1,
                 head_units_value: UnitsValuesType = None,
                 input_dim: int = None,
                 features_metadata: dict = None,
                 embedding_dim: int = 32,
                 num_transformer_layers: int = 8,
                 num_attention_heads: int = 8,
                 attention_dropout: float = 0.1,
                 feedforward_dropout: float = 0.05,
                 feedforward_multiplier: int = 4,
                 encode_categorical_values: bool = True,
                 **kwargs):
        head = RegressionHead(num_outputs=num_outputs,
                              units_values=head_units_value,
                              normalization="layer")
        super().__init__(input_dim=input_dim,
                         features_metadata=features_metadata,
                         embedding_dim=embedding_dim,
                         num_transformer_layers=num_transformer_layers,
                         num_attention_heads=num_attention_heads,
                         attention_dropout=attention_dropout,
                         feedforward_dropout=feedforward_dropout,
                         feedforward_multiplier=feedforward_multiplier,
                         encode_categorical_values=encode_categorical_values,
                         head=head,
                         **kwargs)
        self.num_outputs = num_outputs
        self.head_units_value = head_units_value

    def get_config(self):
        config = super().get_config()
        config.update({'num_outputs': self.num_outputs,
                       'head_units_value': self.head_units_value})
        return config



import tensorflow as tf
import numpy as np
from teras.preprocessing.ctgan import (CTGANDataTransformer as _BaseDataTransformer,
                                       CTGANDataSampler as _BaseDataSampler)
from teras.utils import tf_random_choice
from teras.utils.types import FeaturesNamesType


class TVAEDataTransformer(_BaseDataTransformer):
    """
    TVAEDataTransformer class that is exactly similar to the
    CTGANDataTransformer, it just acts as a wrapper for convenience.

    Reference(s):
        https://arxiv.org/abs/1907.00503
        https://github.com/sdv-dev/CTGAN/

    Args:
        categorical_features: ``List[str]``,
            List of categorical features names in the dataset.

        numerical_features: ``List[str]``,
            List of numerical features names in the dataset.

        max_clusters: ``int``, default 10,
            Maximum Number of clusters to use in ``ModeSpecificNormalization``

        std_multiplier: ``int``, default 4,
            Multiplies the standard deviation in the normalization.

        weight_threshold: ``float``, default 0.005,
            The minimum value a component weight can take to be considered a valid component.
            `weights_` under this value will be ignored.
            (Taken from the official implementation.)

        covariance_type: ``str``, default "full",
            Parameter for the ``GaussianMixtureModel`` class of sklearn.

        weight_concentration_prior_type: ``str``, default "dirichlet_process",
            Parameter for the ``GaussianMixtureModel`` class of sklearn

        weight_concentration_prior: ``float``, default 0.001,
            Parameter for the ``GaussianMixtureModel`` class of sklearn.
    """
    def __init__(self,
                 numerical_features: FeaturesNamesType = None,
                 categorical_features: FeaturesNamesType = None,
                 max_clusters: int = 10,
                 std_multiplier: int = 4,
                 weight_threshold: float = 0.005,
                 covariance_type: str = "full",
                 weight_concentration_prior_type: str = "dirichlet_process",
                 weight_concentration_prior: float = 0.001
                 ):
        super().__init__(numerical_features=numerical_features,
                         categorical_features=categorical_features,
                         max_clusters=max_clusters,
                         std_multiplier=std_multiplier,
                         weight_threshold=weight_threshold,
                         covariance_type=covariance_type,
                         weight_concentration_prior_type=weight_concentration_prior_type,
                         weight_concentration_prior=weight_concentration_prior)


class TVAEDataSampler(_BaseDataSampler):
    """
    TVAEDataSampler class for TVAE architecture.
    It subclasses the ``CTGANDataSampler`` class from CTGAN architecture.

    The two classes share much functionality since TVAE and CTGAN
    are proposed in the same paper and almost all preprocessing
    for both is same.
    There are, however, are a few differences in the ``get_dataset``
    and ``generator`` methods, hence this new subclassed class.

    Args:
        batch_size: ``int``, default 512,
            Batch size to use for the dataset.

        categorical_features: ``List[str]``,
            List of categorical features names

        numerical_features: ``List[str]``,
            List of numerical features names

        metadata:
            A dictionary of metadata computed during data transformation.
            You can access it from the ``.get_metadata()`` of ``CTGANDataTransformer`` instance.
    """
    def __init__(self,
                 batch_size: int = 512,
                 categorical_features: FeaturesNamesType = None,
                 numerical_features: FeaturesNamesType = None,
                 metadata: dict = None
                 ):
        super().__init__(batch_size=batch_size,
                         categorical_features=categorical_features,
                         numerical_features=numerical_features,
                         metadata=metadata)

    def get_dataset(self,
                    x_transformed,
                    x_original=None):
        """
        Args:
            x_transformed: Dataset transformed using ``CTGANDataTransformer`` class
            x_original: Original Dataset - a pandas DataFrame.
                It is used for computing categorical values' probabilities
                for later sampling.
        Returns:
            Returns a tensorflow dataset that utilizes the ``generator`` method
            to create batches of data. This way user can just pass the dataset object to the fit
            method of the model and each batch generated will satisfy all out requirements of sampling
        """
        self.num_samples, self.data_dim = x_transformed.shape
        # adapting the approach from the official implementation
        # to sample evenly across the categories to combat imbalance
        row_idx_raw = [x_original.groupby(feature).groups for feature in self.categorical_features]
        self.row_idx_by_categories = tf.ragged.constant([[values.to_list()
                                                          for values in feat.values()]
                                                         for feat in row_idx_raw])

        dataset = tf.data.Dataset.from_generator(
            self.generator,
            output_signature=(tf.TensorSpec(shape=(self.batch_size, tf.shape(x_transformed)[1]), name="data_batch")
                              ),
            args=[x_transformed],
        )
        return dataset

    def generator(self, x_transformed):
        """
        Used to create a tensorflow dataset.

        Args:
            x_transformed: Dataset transformed by the ``TVAEDataTransformer`` class.

        Returns:
            A batch of data
        """
        # This random_feature_indices variable is required during the sample_cond vector method
        # but since we're using `generator` function to create out tensorflow dataset, this
        # gets called first to generate a batch, so keep in mind that this is where this
        # variable gets its values. We could alternatively just return these indices
        # and pass them as argument to the sample cond_vec but for now let's just work with it.
        num_steps_per_epoch = self.num_samples // self.batch_size
        for _ in range(num_steps_per_epoch):
            random_features_idx = tf.random.uniform([self.batch_size], minval=0,
                                                    maxval=len(self.categorical_features),
                                                    dtype=tf.int32)
            # NOTE: We've precomputed the probabilities in the DataTransformer class for each feature already
            # to speed things up.
            random_features_categories_probs = tf.gather(tf.ragged.constant(self.metadata["categorical"]["categories_probs_all"]),
                                                         indices=random_features_idx)
            random_values_idx = [tf_random_choice(np.arange(len(feature_probs)),
                                                  n_samples=1,
                                                  p=feature_probs)
                                 for feature_probs in random_features_categories_probs]
            random_values_idx = tf.cast(tf.squeeze(random_values_idx), dtype=tf.int32)
            # the official implementation uses actual indices during the sample_cond_vector method
            # but uses the shuffled version in sampling data, so we're gonna do just that.
            shuffled_idx = tf.random.shuffle(tf.range(self.batch_size))
            shuffled_features_idx = tf.gather(random_features_idx, indices=shuffled_idx)
            shuffled_values_idx = tf.gather(random_values_idx, indices=shuffled_idx)
            sample_idx = []
            # TODO make it more efficient -- maybe replace with gather or gathernd or something
            for feat_id, val_id in zip(shuffled_features_idx, shuffled_values_idx):
                s_id = tf_random_choice(self.row_idx_by_categories[tf.squeeze(feat_id)][tf.squeeze(val_id)],
                                        n_samples=1)
                sample_idx.append(tf.squeeze(s_id))

            yield x_transformed[sample_idx]



import tensorflow as tf
from teras.layers.gain.gain_discriminator_block import GAINDiscriminatorBlock


def test_gain_discriminator_block_valid_call():
    gen_block = GAINDiscriminatorBlock(units=32)
    inputs = tf.ones((8, 5), dtype=tf.float32)
    outputs = gen_block(inputs)


def test_gain_discriminator_block_output_shape():
    gen_block = GAINDiscriminatorBlock(units=32)
    inputs = tf.ones((8, 5), dtype=tf.float32)
    outputs = gen_block(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 32





import tensorflow as tf
from teras.layers.regularization import MixUp


def test_mixup_valid_call():
    mixup = MixUp()
    inputs = tf.ones((8, 4), dtype=tf.float32)
    outputs = mixup(inputs)


def test_mixup_output_shape():
    mixup = MixUp()
    inputs = tf.ones((8, 4), dtype=tf.float32)
    outputs = mixup(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 4




import tensorflow as tf
from tensorflow import keras


class VimeSelfSupervisedLoss(keras.losses.Loss):
    """Self supervised loss as proposed by Jinsung Yoon et al.
    in the paper "VIME: Extending the Success of Self- and
    Semi-supervised Learning to Tabular Domain"

    Reference(s):
        https://proceedings.neurips.cc/paper/2020/hash/7d97667a3e056acab9aaf653807b4a03-Abstract.html
    """
    def call(self, y_true=None, y_pred=None):
        """
        Args:
            y_pred: yv_hat_logit
        """
        yu_loss = tf.reduce_mean(tf.nn.moments(y_pred, axes=0)[1])
        return yu_loss


import tensorflow as tf
import tensorflow_probability as tfp
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
import numpy as np


class PiecewiseLinearEncoding:
    """
    PiecewiseLinearEncoding class for encoding numerical features
    as proposed Yury Gorishniy et al. in the paper
    On Embeddings for Numerical Features in Tabular Deep Learning, 2022.

    We implement it like those encoding classes in sklearn.
    It expects the input dataset to be a pandas dataframe.

    Reference(s):
        https://arxiv.org/abs/2203.05556

    Args:
        X: input dataset
        y: target feature or feature name.
            must be specified in case of supervised (tree-based) encoding
        method: quantile, tree
        tree_params: a dictionary of parameters for sklearn's decision tree
        return_type: numpy or tensor.
            numpy returns a numpy array while tensor returns tensorflow tensor
        encoding: piecewise-linear, binary
    """
    def __init__(self,
                 n_bins=10,
                 method="quantile",
                 task="classification",
                 tree_params: dict = {},
                 encoding='piecewise-linear',
                 ):
        self.n_bins = n_bins
        self.method = method
        self.task = task
        self.tree_params = tree_params
        self.encoding = encoding
        self.bin_edges = []

    def fit(self,
            X,
            y=None):
        for feature_idx in range(X.shape[1]):
            feature = X.iloc[:, feature_idx]
            # Discretizing using quantiles
            if self.method.lower() == "quantile":
                feature_n_bins = min(self.n_bins, tf.unique(feature)[0].shape[0])
                quantiles = tf.unique(tfp.stats.quantiles(feature, feature_n_bins))[0]
                self.bin_edges.append(quantiles.numpy())

            elif self.method.lower() == "tree":
                assert y is not None, ("For tree based encoding, y must not be None. Given y: ", y)
                if isinstance(y, str):
                    # y is a feature name that must be present in the feature matrix X
                    # otherwise it's assumed to be a feature itself
                    y = X[y]
                if self.task == "regression":
                    decision_tree = DecisionTreeRegressor(max_leaf_nodes=self.n_bins,
                                                          **self.tree_params)
                else:
                    decision_tree = DecisionTreeClassifier(max_leaf_nodes=self.n_bins,
                                                           **self.tree_params)
                # decision_tree = decision_tree.fit(feature.reshape(-1, 1), y).tree_
                decision_tree = decision_tree.fit(feature.values.reshape(-1, 1), y).tree_
                tree_thresholds = []
                for node_id in range(decision_tree.node_count):
                    # the following condition is True only for split nodes
                    # See https://scikit-learn.org/stable/auto_examples/tree/plot_unveil_tree_structure.html#sphx-glr-auto-examples-tree-plot-unveil-tree-structure-py
                    if decision_tree.children_left[node_id] != decision_tree.children_right[node_id]:
                        tree_thresholds.append(decision_tree.threshold[node_id])
                    tree_thresholds.append(feature.min())
                    tree_thresholds.append(feature.max())
                    self.bin_edges.append(np.array(sorted(set(tree_thresholds))))

    def transform(self,
                  X):
        """Transforms a given dataset based on the encodings learned in fit"""
        self.bins = []
        self.bins_values = []
        for feature_idx in range(X.shape[1]):
            feature = X.iloc[:, feature_idx]
            feature_bin_edges = np.r_[-np.inf, self.bin_edges[feature_idx][1:-1], np.inf]
            self.bins.append(
                tf.cast(
                    tf.searchsorted(
                        feature_bin_edges,
                        feature,
                        side="right"
                        ),
                    dtype="int32").numpy() - 1
            )

            if self.encoding == 'binary':
                self.bins_values.append(
                    tf.ones_like(
                        feature
                    )
                )
            else:
                # encoding must be piecewise-linear
                feature_bin_sizes = self.bin_edges[feature_idx][1:] - self.bin_edges[feature_idx][:-1]
                feature_bins = self.bins[feature_idx]

                self.bins_values.append(
                    (feature - self.bin_edges[feature_idx][feature_bins]) / feature_bin_sizes[feature_bins]
                )

        # n_bins = len(self.bin_edges) - 1
        # bins = tf.stack(self.bins, axis=1)
        bins_values = tf.stack(self.bins_values, axis=1)
        return bins_values.numpy()

    def fit_transform(self, X, y=None):
        self.fit(X, y=y)
        return self.transform(X)


import tensorflow as tf
from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers.node")
class NodeFeatureSelector(keras.layers.Layer):
    """
    NodeFeatureSelector layer for the ``NODE`` architecture.
    This is not part of the official architecture,
    it offers very basic functionality of selecting features
    in case the user specifies the ``max_features`` argument in
    the NODE architecture.
    Strictly speaking, it isn't really needed, but since
    in Teras, we're following a functional approach for our
    models and avoiding the pure subclassing where one
    overrides the ``call`` method. Hence, **our** user friendly
    functional approach necessitates the creation of a layer
    that can handle the extra functionality of feature selection
    in the ``call`` method without modifying the other ``NODE``
    architecture-specific layers such as ``ObliviousDecisionTree``

    Args:
        data_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the input dataset.

        max_features: ``int``, default None,
            Maximum number of features to use.
            If None, all features in the input dataset will be used,
            and the ``FeatureSelector`` layer returns inputs as is.
    """
    def __init__(self,
                 data_dim: int,
                 max_features: int = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.data_dim = data_dim
        self.max_features = max_features

    def call(self, inputs):
        x = inputs
        initial_features = self.data_dim
        if self.max_features is None:
            return x
        tail_features = min(self.max_features, x.shape[-1]) - initial_features
        if tail_features != 0:
            x = tf.concat([x[..., :initial_features], x[..., -tail_features:]],
                          axis=-1)
        return x

    def get_config(self):
        config = super().get_config()
        config.update({"data_dim": self.data_dim,
                       "max_features": self.max_features}
                      )
        return config

    @classmethod
    def from_config(cls, config):
        data_dim = config.pop("data_dim")
        return cls(data_dim=data_dim,
                   **config)
